import time


def watchFuturesLimitTrigger(
    gate, symbol, orderId, doPutTpSl, cancelIfNotOpened, params
):
    if doPutTpSl:
        if (
            "tpSlOrderSide" not in params.keys()
            or "stopLoss" not in params.keys()
            or "takeProfit" not in params.keys()
        ):
            raise ValueError(
                "Must specify 'tpSlOrderSide' and 'stopLoss' and 'takeProfit'"
            )

    if cancelIfNotOpened:
        if "cancelDelaySec" not in params.keys():
            raise ValueError("Must specify 'cancelDelaySec'")
        delayTimeSec = float(params["cancelDelaySec"])
        startDelayTime = time.time()

    print("Watching order")
    while True:
        time.sleep(0.1)
        order = gate.get_order(symbol=symbol, order_id=orderId, futures=True)

        if cancelIfNotOpened:
            if time.time() - startDelayTime > delayTimeSec:
                gate.cancel_order(symbol=symbol, order_id=orderId, futures=True)
                break

        if order["status"] == "NEW":
            continue
        elif order["status"] == "FILLED":
            if doPutTpSl:
                orderSide = params["tpSlOrderSide"]
                stopLoss = params["stopLoss"]
                takeProfit = params["takeProfit"]

                stopLossOrder = gate.create_and_test_futures_order(
                    symbol,
                    orderSide,
                    "STOP_MARKET",
                    stop_price=str(stopLoss),
                    close_position=True,
                    price_protect=True,
                    working_type="MARK_PRICE",
                    time_in_force="GTC",
                )

                takeProfitOrder = gate.create_and_test_futures_order(
                    symbol,
                    orderSide,
                    "TAKE_PROFIT_MARKET",
                    stop_price=str(takeProfit),
                    close_position=True,
                    price_protect=True,
                    working_type="MARK_PRICE",
                    time_in_force="GTC",
                )
                result = gate.make_batch_futures_order([stopLossOrder, takeProfitOrder])
                # print(result)
                break
        elif order["status"] == "CANCELED":
            break

    print("Watching position")
    while True:
        time.sleep(0.1)
        position = gate.get_position_info(symbol)[0]

        if float(position["entryPrice"]) == 0.0:
            gate.cancel_all_symbol_open_orders(symbol, futures=True)
            break


# if __name__ == '__main__':
#     print(sys.argv)
#     time.sleep(3)
#     with open('helloWorld.txt')



from abc import ABC, abstractmethod


class BaseExchange(ABC):
    @abstractmethod
    def __init__(self, credentials, sandbox=False, unified_in_outs=True):
        pass

    @abstractmethod
    def get_balance(self, asset="", futures=False):
        pass

    @abstractmethod
    def symbol_account_trade_history(
        self, symbol, futures=False, from_id=None, limit=None
    ):
        pass

    @abstractmethod
    def test_spot_order(self, order_data):
        pass

    @abstractmethod
    def make_spot_order(self, order_data):
        pass

    @abstractmethod
    def create_and_test_spot_order(
        self,
        symbol,
        side,
        order_type,
        quantity=None,
        price=None,
        time_in_force=None,
        stop_price=None,
        iceberg_qty=None,
        new_order_resp_type=None,
        new_client_order_id=None,
        extra_params=None,
    ):
        pass

    @abstractmethod
    def get_symbol_orders(
        self,
        symbol,
        futures=False,
        order_id=None,
        start_time=None,
        end_time=None,
        limit=None,
    ):
        pass

    @abstractmethod
    def get_open_orders(self, symbol, futures=False):
        pass

    @abstractmethod
    def cancel_all_symbol_open_orders(self, symbol, futures=False):
        pass

    @abstractmethod
    def cancel_order(self, symbol, order_id=None, local_order_id=None, futures=False):
        pass

    @abstractmethod
    def get_order(self, symbol, order_id=None, local_order_id=None, futures=False):
        pass

    @abstractmethod
    def get_trading_fees(self, futures=False):
        pass

    @abstractmethod
    def get_symbol_ticker_price(self, symbol, futures=False):
        pass

    @abstractmethod
    def get_symbol_klines(
        self,
        symbol,
        interval,
        start_time=None,
        end_time=None,
        limit=None,
        futures=False,
        blvtnav=False,
        convert_date_time=False,
        do_clean=False,
        to_clean_dataframe=False,
    ):
        pass

    @abstractmethod
    def get_exchange_time(self, futures=False):
        pass

    @abstractmethod
    def get_symbol_24h_ticker(self, symbol):
        pass

    @abstractmethod
    def test_futures_order(self, futures_order_data):
        pass

    @abstractmethod
    def make_futures_order(self, futures_order_data):
        pass

    @abstractmethod
    def create_and_test_futures_order(
        self,
        symbol,
        side,
        order_type,
        position_side=None,
        time_in_force=None,
        quantity=None,
        reduce_only=None,
        price=None,
        new_client_order_id=None,
        stop_price=None,
        close_position=None,
        activation_price=None,
        callback_rate=None,
        working_type=None,
        price_protect=None,
        new_order_resp_type=None,
        recv_window=None,
        extra_params=None,
    ):
        pass

    @abstractmethod
    def make_batch_futures_order(self, futures_order_datas):
        pass

    @abstractmethod
    def change_initial_leverage(self, symbol, leverage):
        pass

    @abstractmethod
    def change_margin_type(self, symbol, margin_type, params):
        pass

    @abstractmethod
    def change_position_margin(self, symbol, amount):
        pass

    @abstractmethod
    def get_position(self):
        pass

    @abstractmethod
    def spot_best_bid_asks(self, symbol):
        pass

    @abstractmethod
    def get_symbol_order_book(self, symbol, limit=None, futures=False):
        pass

    @abstractmethod
    def get_symbol_recent_trades(self, symbol, limit=None, futures=False):
        pass

    @abstractmethod
    def get_position_info(self, symbol=None):
        pass

    @abstractmethod
    def get_symbol_min_trade(self, symbol, futures=False):
        pass

    @abstractmethod
    def make_sl_tp_limit_futures_order(
        self,
        symbol,
        order_side,
        quantity=None,
        quote_quantity=None,
        enter_price=None,
        take_profit=None,
        stop_loss=None,
        leverage=None,
        margin_type=None,
    ):
        pass

    @abstractmethod
    def get_long_short_ratios(
        self, symbol, period, limit=None, start_time=None, end_time=None
    ):
        pass

    @abstractmethod
    def get_deposit_address(self, coin, network=None):
        pass

    @abstractmethod
    def withdraw(self, coin, address, amount, extra_data):
        pass



class AggregateTradeEvent:

    def __init__(self):
        self.eventType = ""
        self.eventTime = 0
        self.symbol = ""
        self.id = None
        self.price = 0.0
        self.qty = 0.0
        self.firstId = None
        self.lastId = None
        self.time = 0
        self.isBuyerMaker = False

    @staticmethod
    def json_parse(json_wrapper):
        result = AggregateTradeEvent()
        result.eventType = json_wrapper.get_string("e")
        result.eventTime = json_wrapper.get_int("E")
        result.symbol = json_wrapper.get_string("s")
        result.id = json_wrapper.get_int("a")
        result.price = json_wrapper.get_float("p")
        result.qty = json_wrapper.get_float("q")
        result.firstId = json_wrapper.get_int("f")
        result.lastId = json_wrapper.get_int("l")
        result.time = json_wrapper.get_int("T")
        result.isBuyerMaker = json_wrapper.get_boolean("m")
        return result


import contextlib
import logging
import time
from datetime import datetime

import pandas as pd
from binance.spot import Spot

from Exchanges.BaseExchange import BaseExchange
from Utils import BinanceHelpers, DataHelpers
from binance_f import RequestClient
from binance_f.exception.binanceapiexception import BinanceApiException
from binance_f.model.balance import Balance


def is_symbol_status_valid(symbol_name, symbol_datas, futures=False):
    for symbol_data in symbol_datas:
        return bool(
            (
                futures
                and symbol_data.symbol == symbol_name
                and symbol_data.status == "TRADING"
                or not futures
                and symbol_data["symbol"] == symbol_name
                and symbol_data["status"] == "TRADING"
            )
        )


def is_order_data_valid(order: DataHelpers.OrderData):
    check_spot_order_type(order)
    check_order_side(order)
    check_spot_order_response_type(order)
    check_spot_order_time_in_force(order)

    if order.order_type == "LIMIT":
        check_limit_order_data_validity(order)

    elif order.order_type == "MARKET":
        check_spot_market_order_data_validity(order)

    elif order.order_type in ["STOP_LOSS", "TAKE_PROFIT"]:
        check_spot_stop_market_data_validity(order)

    elif order.order_type in ["STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"]:
        check_spot_stop_limit_order_data_validity(order)

    elif order.order_type == "LIMIT_MAKER":
        check_spot_limit_maker_order_data_validity(order)


def check_spot_limit_maker_order_data_validity(order):
    if order.quantity is None or order.price is None:
        raise ValueError("For LIMIT_MAKER order quantity and price must be specified.")


def check_spot_stop_limit_order_data_validity(order):
    if (
        order.time_in_force is None
        or order.quantity is None
        or order.price is None
        or order.stop_price is None
    ):
        raise ValueError(
            "For STOP_LOSS_LIMIT or TAKE_PROFIT_LIMIT orders timeInForce, quantity, price and stopPrice must be specified."
        )


def check_spot_stop_market_data_validity(order):
    if order.quantity is None or order.stop_price is None:
        raise ValueError(
            "For STOP_LOSS or TAKE_PROFIT orders quantity and stopPrice must be specified."
        )


def check_spot_market_order_data_validity(order):
    if order.quantity is None and order.quote_order_qty is None:
        raise ValueError(
            "For MARKET order either quantity or quoteOrderQty must be specified."
        )


def check_limit_order_data_validity(order):
    if order.time_in_force is None or order.quantity is None or order.price is None:
        raise ValueError(
            "For LIMIT order timeInForce, quantity and price must be specified"
        )


def check_spot_order_time_in_force(order):
    if order.time_in_force not in [None, "GTC", "IOC", "FOK"]:
        raise ValueError("Order time in force is not valid.")


def check_spot_order_response_type(order):
    if order.new_order_resp_type not in [None, "ACK", "RESULT", "FULL"]:
        raise ValueError("Order response type is not valid.")


def check_spot_order_type(order):
    if order.order_type not in BinanceExchange.spot_order_types:
        raise ValueError("Order type is not valid.")


def is_futures_order_data_valid(order: DataHelpers.FuturesOrderData):
    check_order_side(order)
    check_futures_order_type(order)
    check_futures_order_side(order)
    check_futures_order_time_in_force(order)
    check_futures_order_working_type(order)
    check_futures_order_response_type(order)
    check_futures_order_close_position(order)
    check_futures_order_callback_rate(order)
    check_futures_order_price_protect(order)
    check_simultaneous_close_position_and_quantity(order)
    check_futures_order_reduce_only(order)
    check_simultaneous_close_position_and_reduce_only(order)

    if order.orderType == "LIMIT":
        check_limit_order_data_validity(order)

    elif order.orderType == "MARKET":
        check_futures_market_order_data_validity(order)

    elif order.orderType in ["STOP", "TAKE_PROFIT"]:
        check_futures_stop_limit_order_data_validity(order)

    elif order.orderType in ["STOP_MARKET", "TAKE_PROFIT_MARKET"]:
        check_futures_stop_market_order_data_validity(order)

    elif order.orderType == "TRAILING_STOP_MARKET":
        check_futures_trailing_order_data_validity(order)


def check_futures_trailing_order_data_validity(order):
    if order.callback_rate is None:
        raise ValueError(
            "For futures TRAILING_STOP_MARKET orders callbackRate must be specified."
        )


def check_futures_stop_market_order_data_validity(order):
    if order.stop_price is None:
        raise ValueError(
            "For futures STOP_MARKET and TAKE_PROFIT_MARKET orders stopPrice must be specified"
        )


def check_futures_stop_limit_order_data_validity(order):
    if order.quantity is None or order.price is None or order.stop_price is None:
        raise ValueError(
            "For futures STOP or TAKE_PROFIT order quantity, price and stopPrice must be specified."
        )


def check_futures_market_order_data_validity(order):
    if order.quantity is None:
        raise ValueError("For MARKET order quantity must be specified.")


def check_simultaneous_close_position_and_reduce_only(order):
    if order.close_position is True and order.reduce_only is True:
        raise ValueError("Can't set both closePosition and reduceOnly to True.")


def check_futures_order_reduce_only(order):
    if order.reduce_only not in [None, True, False]:
        raise ValueError("Futures order reduceOnly is not valid.")


def check_simultaneous_close_position_and_quantity(order):
    if order.close_position is True and order.quantity is not None:
        raise ValueError("Must not specify quantity if closePosition is True.")


def check_futures_order_price_protect(order):
    if order.price_protect not in [None, True, False]:
        raise ValueError("Futures order priceProtect is not valid.")


def check_futures_order_callback_rate(order):
    if order.callback_rate is not None and not (0.1 <= order.callback_rate <= 5):
        raise ValueError("Futures order callbackRate is not valid.")


def check_futures_order_close_position(order):
    if order.close_position not in [None, True, False]:
        raise ValueError("Futures order closePosition is not valid.")


def check_futures_order_response_type(order):
    if order.new_order_resp_type not in [None, "ACK", "RESULT"]:
        raise ValueError("Futures order newOrderRespType is not valid.")


def check_futures_order_working_type(order):
    if order.working_type not in [None, "MARK_PRICE", "CONTRACT_PRICE"]:
        raise ValueError("Futures order workingType is not valid.")


def check_futures_order_time_in_force(order):
    if order.time_in_force not in [None, "GTC", "IOC", "FOK", "GTX"]:
        raise ValueError("Futures order timeInForce is not valid.")


def check_futures_order_side(order):
    if order.position_side not in [None, "BOTH", "LONG", "SHORT"]:
        raise ValueError("Futures order side is not valid")


def check_futures_order_type(order):
    if order.order_type not in BinanceExchange.futures_order_types:
        raise ValueError("Futures order type is not valid.")


def check_order_side(order):
    if order.side not in ["BUY", "SELL"]:
        raise ValueError("Order side can only be 'BUY' or 'SELL'")


class BinanceExchange(BaseExchange):
    time_intervals = [
        "1m",
        "3m",
        "5m",
        "15m",
        "30m",
        "1h",
        "2h",
        "4h",
        "6h",
        "8h",
        "12h",
        "1d",
        "3d",
        "1w",
        "1M",
    ]

    time_indexes_in_candle_data = [0, 6]
    desired_candle_data_indexes = [0, 1, 2, 3, 4, 5, 6, 8]

    spot_order_types = [
        "LIMIT",
        "MARKET",
        "STOP_LOSS",
        "STOP_LOSS_LIMIT",
        "TAKE_PROFIT",
        "TAKE_PROFIT_LIMIT",
        "LIMIT_MAKER",
    ]

    futures_order_types = [
        "LIMIT",
        "MARKET",
        "STOP",
        "STOP_MARKET",
        "TAKE_PROFIT",
        "TAKE_PROFIT_MARKET",
        "TRAILING_STOP_MARKET",
    ]

    def __init__(self, credentials, sandbox=False, unified_in_outs=True):
        self.credentials = credentials
        self.sandbox = sandbox
        self.unified_in_outs = unified_in_outs

        if sandbox:
            self.client = Spot(
                credentials["spot"]["key"],
                credentials["spot"]["secret"],
                base_url="https://testnet.binance.vision",
            )
            self.futures_client = RequestClient(
                api_key=credentials["futures"]["key"],
                secret_key=credentials["futures"]["secret"],
                url="https://testnet.binancefuture.com",
            )
        else:
            self.client = Spot(
                credentials["spot"]["key"], credentials["spot"]["secret"]
            )
            self.futures_client = RequestClient(
                api_key=credentials["futures"]["key"],
                secret_key=credentials["futures"]["secret"],
                url="https://fapi.binance.com",
            )

        self.sub_future_client = None

    def get_balance(self, asset="", futures=False):
        if not futures:
            try:
                balances = self.client.account()["balances"]
            except Exception as e:
                print(e)
                return None
            if asset == "" or asset is None:
                return balances
            return next(
                (balance for balance in balances if balance["asset"] == asset),
                None,
            )
        else:
            balances = [
                balance.toDict() for balance in self.futures_client.get_balance()
            ]
            if asset == "" or asset is None:
                return balances
            for balance in balances:
                if balance["asset"] == asset:
                    return balance
            return Balance.makeFreeBalance(asset)

    def symbol_account_trade_history(
        self, symbol, futures=False, from_id=None, limit=None
    ):
        try:
            if not futures:
                return self.client.my_trades(symbol, fromId=from_id, limit=limit)
            else:
                return [
                    trade.toDict()
                    for trade in self.futures_client.get_account_trades(
                        symbol=symbol, fromId=from_id, limit=limit
                    )
                ]
        except Exception as e:
            print(e)
            return None

    def test_spot_order(self, order_data):
        # sourcery skip: no-conditionals-in-tests
        is_order_data_valid(order_data)

        order_data.set_timestamp()
        params = BinanceHelpers.get_spot_order_as_dict(order_data)

        return self.client.new_order_test(**params)

    def make_spot_order(self, order_data):
        params = BinanceHelpers.get_spot_order_as_dict(order_data)

        response = self.client.new_order(**params)
        logging.info(response)
        return response

    def create_and_test_spot_order(
        self,
        symbol,
        side,
        order_type,
        quantity=None,
        quote_order_qty=None,
        price=None,
        time_in_force=None,
        stop_price=None,
        iceberg_qty=None,
        new_order_resp_type=None,
        new_client_order_id=None,
        extra_params=None,
    ):
        curr_order = DataHelpers.set_spot_order_data(
            iceberg_qty,
            new_client_order_id,
            new_order_resp_type,
            order_type,
            price,
            quantity,
            quote_order_qty,
            side,
            stop_price,
            symbol,
            time_in_force,
            extra_params,
        )

        self.test_spot_order(curr_order)

        return curr_order

    def get_symbol_orders(
        self,
        symbol,
        futures=False,
        order_id=None,
        start_time=None,
        end_time=None,
        limit=None,
    ):
        if not futures:
            return self.client.get_orders(
                symbol=symbol,
                orderId=order_id,
                startTime=start_time,
                endTime=end_time,
                limit=limit,
                timestamp=time.time(),
            )
        else:
            return [
                order.toDict()
                for order in self.futures_client.get_all_orders(
                    symbol,
                    orderId=order_id,
                    startTime=start_time,
                    endTime=end_time,
                    limit=limit,
                )
            ]

    def get_open_orders(self, symbol, futures=False):
        try:
            if not futures:
                return self.client.get_open_orders(symbol, timestamp=time.time())
            else:
                return [
                    order.toDict()
                    for order in self.futures_client.get_open_orders(symbol=symbol)
                ]
        except Exception as e:
            print(e)
            return None

    def cancel_all_symbol_open_orders(self, symbol, futures=False):
        if not futures:
            open_orders = self.get_open_orders(symbol)
            if len(open_orders) == 0:
                return []
            else:
                return self.client.cancel_open_orders(symbol, timestamp=time.time())
        else:
            open_orders = self.get_open_orders(symbol, futures=True)

            if len(open_orders) == 0:
                return []
            order_ids = [order["orderId"] for order in open_orders]

            return [
                res.toDict()
                for res in self.futures_client.cancel_list_orders(
                    symbol=symbol, orderIdList=order_ids
                )
            ]

    def cancel_order(self, symbol, order_id=None, local_order_id=None, futures=False):
        error_message = (
            "Specify either order Id in the exchange or local Id sent with the order"
        )
        if futures:
            if order_id is not None:
                return self.futures_client.cancel_order(
                    symbol, orderId=order_id
                ).toDict()
            elif local_order_id is not None:
                return self.futures_client.cancel_order(
                    symbol, origClientOrderId=local_order_id
                ).toDict()
            else:
                raise ValueError(error_message)

        elif order_id is not None:
            return self.client.cancel_order(
                symbol, orderId=order_id, timestamp=time.time()
            )
        elif local_order_id is not None:
            return self.client.cancel_order(
                symbol, origClientOrderId=local_order_id, timestamp=time.time()
            )
        else:
            raise ValueError(error_message)

    def get_order(self, symbol, order_id=None, local_order_id=None, futures=False):
        error_message = (
            "Specify either order Id in the exchange or local Id sent with the order"
        )
        if futures:
            if order_id is not None:
                return self.futures_client.get_order(symbol, orderId=order_id).toDict()
            elif local_order_id is not None:
                return self.futures_client.get_order(
                    symbol, origClientOrderId=local_order_id
                ).toDict()
            else:
                raise ValueError(error_message)

        elif order_id is not None:
            return self.client.get_order(
                symbol, orderId=order_id, timestamp=time.time()
            )
        elif local_order_id is not None:
            return self.client.get_order(
                symbol, origClientOrderId=local_order_id, timestamp=time.time()
            )
        else:
            raise ValueError(error_message)

    def get_trading_fees(self, symbol=None, futures=False):
        if symbol:
            return self.client.trade_fee(symbol=symbol)[0]
        else:
            return self.client.trade_fee(symbol=symbol)

    def get_symbol_ticker_price(self, symbol, futures=False):
        if futures:
            return self.futures_client.get_symbol_price_ticker(symbol=symbol)[0].price
        else:
            return float(self.client.ticker_price(symbol)["price"])

    def get_symbol_klines(
        self,
        symbol,
        interval,
        start_time=None,
        end_time=None,
        limit=None,
        futures=False,
        blvtnav=False,
        convert_date_time=False,
        do_clean=False,
        to_clean_dataframe=False,
    ):
        if interval not in BinanceExchange.time_intervals:
            raise ValueError("Time interval is not valid.")

        if futures:
            data = self._get_futures_symbol_klines(
                blvtnav, end_time, interval, limit, start_time, symbol
            )
        else:
            data = self._get_spot_symbol_klines(
                end_time, interval, limit, start_time, symbol
            )

        if convert_date_time or to_clean_dataframe:
            BinanceHelpers.klines_convert_date(data, self.time_indexes_in_candle_data)

        if not do_clean and not to_clean_dataframe:
            return data

        final_data_array = BinanceHelpers.get_klines_desired_only_cols(
            data, self.desired_candle_data_indexes
        )

        if to_clean_dataframe:
            return BinanceHelpers.klines_convert_to_pandas(final_data_array)
        return final_data_array

    def _get_spot_symbol_klines(self, end_time, interval, limit, start_time, symbol):
        data = self.client.klines(
            symbol, interval, startTime=start_time, endTime=end_time, limit=limit
        )
        for datum in data:
            for idx in range(len(datum)):
                if idx in BinanceExchange.time_indexes_in_candle_data:
                    continue
                datum[idx] = float(datum[idx])
        return data

    def _get_futures_symbol_klines(
        self, blvtnav, end_time, interval, limit, start_time, symbol
    ):
        if blvtnav:
            candles = self.futures_client.get_blvt_nav_candlestick_data(
                symbol=symbol,
                interval=interval,
                startTime=start_time,
                endTime=end_time,
                limit=limit,
            )
        else:
            candles = self.futures_client.get_candlestick_data(
                symbol=symbol,
                interval=interval,
                startTime=start_time,
                endTime=end_time,
                limit=limit,
            )
        return [candle.toArray() for candle in candles]

    def get_exchange_time(self, futures=False):
        try:
            if not futures:
                return self.client.time()["serverTime"]
            else:
                return self.futures_client.get_servertime()
        except Exception as e:
            print(e)
            return None

    def get_symbol_24h_ticker(self, symbol):
        try:
            return self.client.ticker_24hr(symbol)
        except Exception as e:
            print(e)
            return None

    def test_futures_order(self, futures_order_data):
        # sourcery skip: no-conditionals-in-tests
        if not is_futures_order_data_valid(futures_order_data):
            raise ValueError("Incomplete data provided.")
        return futures_order_data

    def make_futures_order(self, futures_order_data):
        params = BinanceHelpers.get_futures_order_as_dict(futures_order_data)

        response = self.futures_client.post_order(**params)
        return response.toDict()

    def create_and_test_futures_order(
        self,
        symbol,
        side,
        order_type,
        position_side=None,
        time_in_force=None,
        quantity=None,
        reduce_only=None,
        price=None,
        new_client_order_id=None,
        stop_price=None,
        close_position=None,
        activation_price=None,
        callback_rate=None,
        working_type=None,
        price_protect=None,
        new_order_resp_type=None,
        recv_window=None,
        extra_params=None,
        quote_quantity=None,
    ):
        curr_order = DataHelpers.set_futures_order_data(
            activation_price,
            callback_rate,
            close_position,
            extra_params,
            new_client_order_id,
            new_order_resp_type,
            order_type,
            position_side,
            price,
            price_protect,
            quantity,
            reduce_only,
            side,
            stop_price,
            symbol,
            time_in_force,
            working_type,
            quote_quantity,
        )

        self.test_futures_order(curr_order)

        return curr_order

    def make_batch_futures_order(self, futures_order_datas):
        batch_orders = BinanceHelpers.make_batch_order_data(futures_order_datas)

        order_results = self.futures_client.post_batch_order(batch_orders)

        return [order.toDict() for order in order_results]

    def cancel_all_symbol_futures_orders_with_countdown(self, symbol, countdown_time):
        return self.futures_client.auto_cancel_all_orders(symbol, countdown_time)

    def change_initial_leverage(self, symbol, leverage):
        return self.futures_client.change_initial_leverage(
            symbol=symbol, leverage=leverage
        ).toDict()["leverage"]

    def change_margin_type(self, symbol, margin_type, params=None):
        if margin_type not in ["ISOLATED", "CROSSED"]:
            raise ValueError("Margin type specified is not acceptable")
        with contextlib.suppress(BinanceApiException):
            result = self.futures_client.change_margin_type(
                symbol=symbol, marginType=margin_type
            )
            return result["code"] == 200

    def change_position_margin(self, symbol, amount):
        add_or_sub = 1 if amount >= 0 else 2
        result = self.futures_client.change_position_margin(
            symbol=symbol, amount=amount, type=add_or_sub
        )
        return result["code"] == 200

    def get_position(self):
        return self.futures_client.get_position()

    def spot_best_bid_asks(self, symbol):
        return self.client.book_ticker(symbol=symbol)

    def get_symbol_order_book(self, symbol, limit=None, futures=False):
        if not futures:
            return (
                self.client.depth(symbol)
                if limit is None
                else self.client.depth(symbol, limit=limit)
            )
        if limit is None:
            return self.futures_client.get_order_book(symbol=symbol).toDict()
        else:
            return self.futures_client.get_order_book(
                symbol=symbol, limit=limit
            ).toDict()

    def get_symbol_recent_trades(self, symbol, limit=None, futures=False):
        if limit is not None:
            if limit > 1000:
                limit = 1000
            elif limit < 1:
                limit = 1
        if not futures:
            return (
                pd.DataFrame(self.client.trades(symbol=symbol))
                if limit is None
                else pd.DataFrame(self.client.trades(symbol, limit=limit))
            )
        if limit is None:
            return pd.DataFrame(
                self.futures_client.get_recent_trades_list(symbol=symbol)
            )
        else:
            return pd.DataFrame(
                self.futures_client.get_recent_trades_list(symbol=symbol, limit=limit)
            )

    def get_position_info(self, symbol=None):
        return self.futures_client.get_position_v2(symbol)

    def get_symbol_min_trade(self, symbol, futures=False):
        ticker_price = self.get_symbol_ticker_price(symbol, futures=futures)

        if futures:
            exchange_info = self.futures_client.get_exchange_information()

            for sym in exchange_info.symbols:
                if sym.symbol == symbol:
                    symbol_filters = sym.filters
                    return BinanceHelpers.extract_symbol_info_from_filters(
                        symbol_filters, ticker_price
                    )
        else:
            exchange_info = self.client.exchange_info()

            for sym in exchange_info["symbols"]:
                if sym["symbol"] == symbol:
                    symbol_filters = sym["filters"]
                    return BinanceHelpers.extract_symbol_info_from_filters(
                        symbol_filters, ticker_price
                    )
        return None

    def get_income_history(
        self, symbol, income_type=None, start_time=None, end_time=None, limit=None
    ):
        return self.futures_client.get_income_history(
            symbol=symbol,
            incomeType=income_type,
            startTime=start_time,
            endTime=end_time,
            limit=limit,
        )

    def make_sl_tp_limit_futures_order(
        self,
        symbol,
        order_side,
        quantity=None,
        quote_quantity=None,
        enter_price=None,
        take_profit=None,
        stop_loss=None,
        leverage=None,
        margin_type=None,
    ):
        symbol_info = self.get_symbol_min_trade(symbol=symbol, futures=True)

        quantity = DataHelpers.get_quantity(
            enter_price, quantity, quote_quantity, symbol_info["precisionStep"]
        )
        self._set_leverage(leverage, symbol)
        self.change_margin_type(symbol, margin_type)
        tp_sl_order_side = "BUY" if order_side.upper() == "SELL" else "SELL"

        main_order, stop_loss_order, take_profit_order = self.create_sl_tp_limit_orders(
            enter_price,
            order_side,
            quantity,
            stop_loss,
            symbol,
            take_profit,
            tp_sl_order_side,
        )

        ordering_result = self.make_batch_futures_order(
            [main_order, stop_loss_order, take_profit_order]
        )

        return DataHelpers.get_tp_sl_limit_order_ids(ordering_result)

    def create_sl_tp_limit_orders(
        self,
        enter_price,
        order_side,
        quantity,
        stop_loss,
        symbol,
        take_profit,
        tp_sl_order_side,
    ):
        main_order = self.create_and_test_futures_order(
            symbol,
            order_side.upper(),
            "LIMIT",
            quantity=str(quantity),
            price=str(enter_price),
            time_in_force="GTC",
        )
        stop_loss_order = self.create_and_test_futures_order(
            symbol,
            tp_sl_order_side,
            "STOP_MARKET",
            stop_price=str(stop_loss),
            close_position=True,
            price_protect=True,
            working_type="MARK_PRICE",
            time_in_force="GTC",
        )
        take_profit_order = self.create_and_test_futures_order(
            symbol,
            tp_sl_order_side,
            "TAKE_PROFIT_MARKET",
            stop_price=str(take_profit),
            close_position=True,
            price_protect=True,
            working_type="MARK_PRICE",
            time_in_force="GTC",
        )
        return main_order, stop_loss_order, take_profit_order

    def make_sl_tp_market_futures_order(
        self,
        symbol,
        order_side,
        quantity=None,
        quote_quantity=None,
        take_profit=None,
        stop_loss=None,
        leverage=None,
        margin_type=None,
    ):
        symbol_info = self.get_symbol_min_trade(symbol=symbol, futures=True)
        market_price = self.get_symbol_ticker_price(symbol=symbol, futures=True)

        quantity = DataHelpers.get_quantity(
            market_price, quantity, quote_quantity, symbol_info["precisionStep"]
        )
        self._set_leverage(leverage, symbol)
        self.change_margin_type(symbol, margin_type)
        tp_sl_order_side = "BUY" if order_side.upper() == "SELL" else "SELL"

        has_sl, has_tp, orders_list = self.create_sl_tp_market_orders(
            order_side, quantity, stop_loss, symbol, take_profit, tp_sl_order_side
        )

        ordering_result = self.make_batch_futures_order(orders_list)

        return DataHelpers.get_tp_sl_market_order_ids(
            ordering_result, has_sl=has_sl, has_tp=has_tp
        )

    def create_sl_tp_market_orders(
        self, order_side, quantity, stop_loss, symbol, take_profit, tp_sl_order_side
    ):
        main_order = self.create_and_test_futures_order(
            symbol, order_side.upper(), "MARKET", quantity=str(quantity)
        )
        orders_list = [main_order]
        has_tp = False
        has_sl = False
        if stop_loss is not None:
            stop_loss_order = self.create_and_test_futures_order(
                symbol,
                tp_sl_order_side,
                "STOP_MARKET",
                stop_price=str(stop_loss),
                close_position=True,
                price_protect=True,
                working_type="MARK_PRICE",
                time_in_force="GTC",
            )
            orders_list.append(stop_loss_order)
            has_sl = True
        if take_profit is not None:
            take_profit_order = self.create_and_test_futures_order(
                symbol,
                tp_sl_order_side,
                "TAKE_PROFIT_MARKET",
                stop_price=str(take_profit),
                close_position=True,
                price_protect=True,
                working_type="MARK_PRICE",
                time_in_force="GTC",
            )
            orders_list.append(take_profit_order)
            has_tp = True
        return has_sl, has_tp, orders_list

    def _set_leverage(self, leverage, symbol):
        set_leverage_result = self.change_initial_leverage(symbol, leverage)
        print("Leverage changed.")
        if (
            isinstance(set_leverage_result, dict)
            and set_leverage_result["leverage"] != leverage
            or not isinstance(set_leverage_result, dict)
            and isinstance(set_leverage_result, float)
            and set_leverage_result != leverage
            or not isinstance(set_leverage_result, dict)
            and not isinstance(set_leverage_result, float)
        ):
            raise ConnectionError("Could not change leverage.")

    def get_symbol_list(self, futures=False):
        if futures:
            return [
                symbolInfo.symbol
                for symbolInfo in self.futures_client.get_exchange_information().symbols
                if symbolInfo.status == "TRADING"
            ]

    def get_symbol_24h_changes(self, futures=False):
        symbol_datas = []
        if futures:
            symbol_status = self.futures_client.get_exchange_information().symbols
            symbol_datas.extend(
                (symbolInfo.symbol, symbolInfo.priceChangePercent)
                for symbolInfo in self.futures_client.get_ticker_price_change_statistics()
                if is_symbol_status_valid(
                    symbolInfo.symbol, symbol_status, futures=True
                )
            )
        else:
            symbol_status = self.client.exchange_info()["symbols"]
            symbol_datas.extend(
                (symbolInfo["symbol"], float(symbolInfo["priceChangePercent"]))
                for symbolInfo in self.client.ticker_24hr()
                if is_symbol_status_valid(
                    symbolInfo["symbol"], symbol_status, futures=False
                )
            )
        return sorted(symbol_datas, key=lambda x: x[1], reverse=True)

    def get_latest_symbol_names(self, num_of_symbols=None, futures=False):
        symbol_datas = []
        if not futures:
            raise NotImplementedError("Only available for futures market.")

        for symbolInfo in self.futures_client.get_exchange_information().symbols:
            symbol_datas.append(
                (
                    symbolInfo.symbol,
                    datetime.fromtimestamp(float(symbolInfo.onboardDate) / 1000),
                )
            )
            symbol_datas.sort(key=lambda x: x[1], reverse=True)
        if num_of_symbols is not None and num_of_symbols > len(symbol_datas):
            num_of_symbols = len(symbol_datas)
        return symbol_datas[:num_of_symbols]

    def get_long_short_ratios(
        self, symbol, period, limit=None, start_time=None, end_time=None
    ):
        if limit is None:
            limit = 30
        top_long_short_accounts = self.futures_client.get_top_long_short_accounts(
            symbol=symbol,
            period=period,
            startTime=start_time,
            endTime=end_time,
            limit=limit,
        )
        top_long_short_positions = self.futures_client.get_top_long_short_positions(
            symbol=symbol,
            period=period,
            startTime=start_time,
            endTime=end_time,
            limit=limit,
        )
        long_short_ratio = self.futures_client.get_global_long_short_accounts(
            symbol=symbol,
            period=period,
            startTime=start_time,
            endTime=end_time,
            limit=limit,
        )

        return {
            "topLongShortAccounts": top_long_short_accounts,
            "topLongShortPositions": top_long_short_positions,
            "longShortRatio": long_short_ratio,
        }

    def get_deposit_address(self, coin, network=None):
        return self.client.deposit_address(coin=coin, network=network)

    def withdraw(self, coin, address, amount, **extra_data):
        return self.client.withdraw(
            coin=coin, amount=amount, address=address, **extra_data
        )



class AdlQuantile:

    def __init__(self):
        self.symbol = ""
        self.adlQuantile = ""

    @staticmethod
    def json_parse(json_wrapper):
        result = AdlQuantile()
        result.symbol = json_wrapper.get_string("symbol")
        result.adlQuantile = json_wrapper.get_string("adlQuantile")
        return result





class OpenInterest:

    def __init__(self):
        self.symbol = ""
        self.openInterest = 0.0
    
    @staticmethod
    def json_parse(json_data):
        result = OpenInterest()
        result.symbol = json_data.get_string("symbol")
        result.openInterest = json_data.get_float("openInterest")

        return result



class LiquidationOrder:

    def __init__(self):
        self.symbol = ""
        self.price = 0.0
        self.origQty = 0.0
        self.executedQty = 0.0
        self.averagePrice = 0.0
        self.timeInForce = ""
        self.type = ""
        self.side = ""
        self.time = 0
    
    @staticmethod
    def json_parse(json_data):
        result = LiquidationOrder()
        result.symbol = json_data.get_string("symbol")
        result.price = json_data.get_float("price")
        result.origQty = json_data.get_float("origQty")
        result.executedQty = json_data.get_float("executedQty")
        result.averagePrice = json_data.get_float("averagePrice")
        result.timeInForce = json_data.get_string("timeInForce")
        result.type = json_data.get_string("symbol")
        result.side = json_data.get_string("side")
        result.time = json_data.get_int("time")

        return result



class LiquidationOrder:

    def __init__(self):
        self.symbol = ""
        self.side = ""
        self.type = ""
        self.timeInForce = ""
        self.origQty = 0.0
        self.price = 0.0
        self.averagePrice = 0.0
        self.orderStatus = ""
        self.lastFilledQty = 0.0
        self.lastFilledAccumulatedQty = 0.0
        self.time = 0


class LiquidationOrderEvent:

    def __init__(self):
        self.eventType = ""
        self.eventTime = 0
        self.data = None

    @staticmethod
    def json_parse(json_wrapper):
        result = LiquidationOrderEvent()
        result.eventType = json_wrapper.get_string("e")
        result.eventTime = json_wrapper.get_int("E")
        data = json_wrapper.get_object("o")
        element = LiquidationOrder()
        element.symbol = data.get_string("s")
        element.side = data.get_string("S")
        element.type = data.get_string("o")
        element.timeInForce = data.get_string("f")
        element.origQty = data.get_float("q")
        element.price = data.get_float("p")
        element.averagePrice = data.get_float("ap")
        element.orderStatus = data.get_string("X")
        element.lastFilledQty = data.get_float("l")
        element.lastFilledAccumulatedQty = data.get_float("z")
        element.time = data.get_int("T")
        result.data = element
        return result


class SymbolPrice:

    def __init__(self):
        self.symbol = ""
        self.price = 0.0
        self.time = 0

    @staticmethod
    def json_parse(json_data):
        result = SymbolPrice()
        result.symbol = json_data.get_string("symbol")
        result.price = json_data.get_float("price")
        result.time = json_data.get_int("time")
        return result


class BaseAsset:

    def __init__(self):
        self.baseAsset = ""
        self.weightInQuantity = 0.0
        self.weightInPercentage = 0.0
        
    @staticmethod
    def json_parse(json_data):
        result = BaseAsset()
        result.baseAsset = json_data.get_string("baseAsset")
        result.weightInQuantity = json_data.get_float("weightInQuantity")
        result.weightInPercentage = json_data.get_float("weightInPercentage")
        return result

class IndexInfo:

    def __init__(self):
        self.symbol = ""
        self.time = 0
        self.baseAssetList = list()
    
    @staticmethod
    def json_parse(json_data):
        result = IndexInfo()
        result.symbol = json_data.get_string("symbol")
        result.time = json_data.get_int("time")

        element_list = list()
        data_list = json_data.get_array("baseAssetList")
        for item in data_list.get_items():
            element = BaseAsset.json_parse(item)
            element_list.append(element)
        result.baseAssetList = element_list

        return result



import base64
import hashlib
import hmac
import datetime
from urllib import parse
import urllib.parse
from binance_f.exception.binanceapiexception import BinanceApiException


def create_signature(secret_key, builder):
    if secret_key is None or secret_key == "":
        raise BinanceApiException(BinanceApiException.KEY_MISSING,  "Secret key are required")

#    keys = builder.param_map.keys()
#    query_string = '&'.join(['%s=%s' % (key, parse.quote(builder.param_map[key], safe='')) for key in keys])
    query_string = builder.build_url()
    signature = hmac.new(secret_key.encode(), msg=query_string.encode(), digestmod=hashlib.sha256).hexdigest()
    builder.put_url("signature", signature)


def create_signature_with_query(secret_key, query):
    if secret_key is None or secret_key == "":
        raise BinanceApiException(BinanceApiException.KEY_MISSING,  "Secret key are required")

    signature = hmac.new(secret_key.encode(), msg=query.encode(), digestmod=hashlib.sha256).hexdigest()
    
    return signature


def utc_now():
    return datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S')





class TakerBuySellRatio:

    def __init__(self):
        self.buySellRatio = 0.0
        self.buyVol = 0.0
        self.sellVol = 0.0
        self.timestamp = 0
    
    @staticmethod
    def json_parse(json_data):
        result = TakerBuySellRatio()
        result.buySellRatio = json_data.get_float("buySellRatio")
        result.buyVol = json_data.get_float("buyVol")
        result.sellVol = json_data.get_float("sellVol")
        result.timestamp = json_data.get_int("timestamp")

        return result



"""
TradeGate - An algorithmic trading library to use as a gateway to different exchanges.
"""

__version__ = "0.3.2"

from Exchanges import BinanceExchange, BybitExchange, KuCoinExchange


def getCorrectExchange(exchangeName):
    if exchangeName.lower() == "binance":
        return BinanceExchange.BinanceExchange
    if exchangeName.lower() == "bybit":
        return BybitExchange.BybitExchange
    if exchangeName.lower() == "kucoin":
        return KuCoinExchange.KuCoinExchange


class TradeGate:
    def __init__(self, configDict, sandbox=False):
        self.exchangeName = configDict["exchangeName"]
        exchangeClass = getCorrectExchange(self.exchangeName)
        if sandbox:
            self.apiKey = configDict["credentials"]["test"]["spot"]["key"]
            self.apiSecret = configDict["credentials"]["test"]["spot"]["secret"]

            self.exchange = exchangeClass(
                configDict["credentials"]["test"], sandbox=True
            )
        else:
            self.apiKey = configDict["credentials"]["main"]["spot"]["key"]
            self.apiSecret = configDict["credentials"]["main"]["spot"]["secret"]

            self.exchange = exchangeClass(
                configDict["credentials"]["main"], sandbox=False
            )

    def getBalance(self, asset=None, futures=False):
        """Returns account balance of all assets or a single asset

        :param asset: a valid asset name, defaults to None
        :type asset: str , optional
        :param futures: False for spot market and True for futures market, defaults to False
        :type futures: bool , optional
        :return: Returns a single asset balance or list of assets if no asset was specified.
        :rtype: dict or list(dict)
        :Output with asset specified:

            .. code-block:: python

                {
                    'asset': 'BNB',
                    'free': '1000.00000000',
                    'locked': '0.00000000'
                }

        :Output without asset specified:

            .. code-block:: python

                [
                    {
                        'asset': 'BNB',
                        'free': '1000.00000000',
                        'locked': '0.00000000'
                    },
                    {
                        'asset': 'BTC',
                        'free': '1.02000000',
                        'locked': '0.00000000'
                    },
                    ...
                ]

        """
        return self.exchange.get_balance(asset, futures)

    def createAndTestSpotOrder(
        self,
        symbol,
        side,
        orderType,
        quantity=None,
        quote_order_qty=None,
        price=None,
        timeInForce=None,
        stopPrice=None,
        icebergQty=None,
        newOrderRespType=None,
        newClientOrderId=None,
        extraParams=None,
    ):
        """ Create a OrderData object and test the given parameters for validity. The object returned is then used \
        to send an order to the exchange by :func:`make_spot_order() <TradeGate.TradeGate.make_spot_order>`

        :param symbol: Symbol of the order
        :type symbol: str
        :param side: Side of the order. Either '**BUY**' or '**SELL**'
        :type side: str
        :param orderType: Type of the order. can be '**MARKET**', '**LIMIT**' or others (Check exchange's API documentation)
        :type orderType: str
        :param quantity: The amount of base asset of the order.
        :type quantity: float , optional
        :param quantity: The amount of quote asset of the order.
        :type quantity: float , optional
        :param price: Order price
        :type price: float , optional
        :param timeInForce: Order's time in force. Can be '**GTC**', '**IOC**', '**FOK**' or specific to the exchange. \
        Check the exchange's API documentation for more options.
        :type timeInForce: str , optional
        :param stopPrice: Price condition to trigger the order. Only for stop order type.
        :type stopPrice: float , optional
        :param icebergQty: Only for Binance. Amount of the base asset to be hide on the exchange.
        :type icebergQty: float , optional
        :param newOrderRespType: Only for Binance. Response of the order, either '**ACK**' or '**RESULT**'.
        :type newOrderRespType: str , optional
        :param newClientOrderId: Custom string to identify your order for yourself.
        :type newClientOrderId: str , optional
        :param extraParams: Extra parameters for other exchanges than Binance.
        :type extraParams: dict , optional
        :return: An orderData object of the created order, ready to be submitted using \
        :func:`make_spot_order() <TradeGate.TradeGate.make_spot_order>`
        :rtype: OrderData

        :Notes:

            * Input parameters are based on Binance API. We map the appropriate parameters from input to the exchange \
            requirements, but other parameters must be inside the '**extraParams**' dictionary.
            * Some parameters have values that are only valid for some exchanges. See each exchange's documentation for\
             these values.
            * ValueError exception is raised when wrong combination of parameters are sent. Check the exception's \
            message for guidance.
            * Do not send '**price**' with '**MARKET**' orders.
            * You must specify '**price**' for '**LIMIT**' orders.
            * For **ByBit**:

                * Order type can only be '**MARKET**', '**LIMIT**' or '**LIMIT_MAKER**'

            * For **KuCoin**:

                * To send a stop order, set '**orderType**' to **STOP_MARKET** or **STOP_LIMIT**.
                * Send '**stop**' (either **loss** or **entry**) in the '**extraParams**' when specifying '**stopPrice**'.
                * You can send '**cancelAfter**' parameter inside '**extraParams**' to cancel order after **n** seconds \
                only if '**timeInForce**' value is **GTT**
                * You can send '**postOnly**', '**hidden**', '**iceberg**' (bool) and '**visibleSize**' parameters inside \
                '**extraParams**'.
                * Valid values for the '**timeInForce**' variable: **GTC** (Good Till Cancel), **IOC** \
                (Immediate Or Cancel), **GTT** (Good Till Time) and **FOK** (Fill Or Kill)

        """
        return self.exchange.create_and_test_spot_order(
            symbol,
            side,
            orderType,
            quantity,
            quote_order_qty,
            price,
            timeInForce,
            stopPrice,
            icebergQty,
            newOrderRespType,
            newClientOrderId,
            extraParams,
        )

    def makeSpotOrder(self, orderData):
        """Make a spot order

        :param orderData: OrderData created using :func:`make_spot_order() <TradeGate.TradeGate.make_spot_order>`
        :type orderData: OrderData
        :return: submitted order information
        :rtype: dict
        :Output:

            .. code-block:: python

                {
                    'symbol': 'BTCUSDT',
                    'orderId': 7048832,
                    'orderListId': -1,
                    'clientOrderId': 'Jes2sprMLHxEUXHyJOUEAr',
                    'transactTime': 1656343106986,
                    'price': '20000.00000000',
                    'origQty': '0.00200000',
                    'executedQty': '0.00000000',
                    'cummulativeQuoteQty': '0.00000000',
                    'status': 'NEW',
                    'timeInForce': 'GTC',
                    'type': 'LIMIT',
                    'side': 'BUY',
                    'fills': []
                }

        """
        return self.exchange.make_spot_order(orderData)

    def getSymbolOrders(
        self,
        symbol,
        futures=False,
        orderId=None,
        startTime=None,
        endTime=None,
        limit=None,
    ):
        """Get History of orders submitted

        :param symbol: The order's symbol
        :type symbol: str
        :param orderId: Only return orders from the order with this orderId onwards.
        :type orderId: long, optional
        :param startTime: Timestamp for the start of the data
        :type startTime: long, optional
        :param endTime: Timestamp for the end of the data
        :type endTime: long, optional
        :param limit: Maximum number of order datas to return
        :type limit: int, optional
        :param futures: False for spot market and True for futures market, defaults to False
        :type futures: bool , optional
        :return: A list of orders datas
        :rtype: list(dict)
        :Output:

            .. code-block:: python

                [
                    {
                        'symbol': 'BTCUSDT',
                        'orderId': 7020578,
                        'orderListId': -1,
                        'clientOrderId': 'tI8wxMsbaUk6hcf5QpC8zx',
                        'price': '20000.00000000',
                        'origQty': '0.00200000',
                        'executedQty': '0.00000000',
                        'cummulativeQuoteQty': '0.00000000',
                        'status': 'NEW',
                        'timeInForce': 'GTC',
                        'type': 'LIMIT',
                        'side': 'BUY',
                        'stopPrice': '0.00000000',
                        'icebergQty': '0.00000000',
                        'time': 1656338254295,
                        'updateTime': 1656338254295,
                        'isWorking': True,
                        'origQuoteOrderQty': '0.00000000'
                    },
                    {
                        'symbol': 'BTCUSDT',
                        'orderId': 7048832,
                        'orderListId': -1,
                        'clientOrderId': 'Jes2sprMLHxEUXHyJOUEAr',
                        'price': '20000.00000000',
                        'origQty': '0.00200000',
                        'executedQty': '0.00000000',
                        'cummulativeQuoteQty': '0.00000000',
                        'status': 'NEW',
                        'timeInForce': 'GTC',
                        'type': 'LIMIT',
                        'side': 'BUY',
                        'stopPrice': '0.00000000',
                        'icebergQty': '0.00000000',
                        'time': 1656343106986,
                        'updateTime': 1656343106986,
                        'isWorking': True,
                        'origQuoteOrderQty': '0.00000000'
                    }
                ]

        :Notes:

            * Specify the symbol for better output.

        """
        return self.exchange.get_symbol_orders(
            symbol=symbol,
            futures=futures,
            order_id=orderId,
            start_time=startTime,
            end_time=endTime,
            limit=limit,
        )

    def getOpenOrders(self, symbol, futures=False):
        """Get order datas of all open orders for a symbol

        :param symbol: The symbol
        :type symbol: str
        :param futures: False for spot market and True for futures market, defaults to False
        :type futures: bool , optional
        :return: Order datas of the open orders
        :rtype: list(dict)
        :Output:

            .. code-block:: python

                [
                    {
                        'symbol': 'BTCUSDT',
                        'orderId': 3051997664,
                        'clientOrderId': 'VhxN3SezOKgpvACYe2qhO6',
                        'transactTime': 1656337589565,
                        'price': 20000.0,
                        'origQty': 0.002,
                        'executedQty': 0.0,
                        'cummulativeQuoteQty': 0.0,
                        'status': 'NEW',
                        'timeInForce': 'GTC',
                        'type': 'LIMIT',
                        'side': 'BUY',
                        'extraData':
                        {
                            'reduceOnly': False,
                            'stopPrice': 0.0,
                            'workingType': 'CONTRACT_PRICE',
                            'avgPrice': 0.0,
                            'origType': 'LIMIT',
                            'positionSide': 'BOTH',
                            'activatePrice': None,
                            'priceRate': None,
                            'closePosition': False
                        }
                    },
                    {
                        'symbol': 'BTCUSDT',
                        'orderId': 3051997723,
                        'clientOrderId': 'v1nT9fQjDGVRqHBtDYotTh',
                        'transactTime': 1656337600701,
                        'price': 20000.0,
                        'origQty': 0.002,
                        'executedQty': 0.0,
                        'cummulativeQuoteQty': 0.0,
                        'status': 'NEW',
                        'timeInForce': 'GTC',
                        'type': 'LIMIT',
                        'side': 'BUY',
                        'extraData':
                        {
                            'reduceOnly': False,
                            'stopPrice': 0.0,
                            'workingType':
                            'CONTRACT_PRICE',
                            'avgPrice': 0.0,
                            'origType': 'LIMIT',
                            'positionSide': 'BOTH',
                            'activatePrice': None,
                            'priceRate': None,
                            'closePosition': False
                        }
                    }
                ]

        """
        return self.exchange.get_open_orders(symbol, futures)

    def getOrder(self, symbol, orderId=None, localOrderId=None, futures=False):
        """Get an order's data

        :param symbol: The Order's symbol
        :type symbol: str
        :param orderId: Exchange assigned '**orderId**' of the order.
        :type orderId: long, optional
        :param localOrderId: User assigned '**clientOrderId**' of the order.
        :type localOrderId: long, optional
        :param futures: False for spot market and True for futures market, defaults to False
        :type futures: bool , optional
        :return: The order data
        :rtype: dict
        :Output:

            .. code-block:: python

                {
                    'symbol': 'BTCUSDT',
                    'orderId': 3051988035,
                    'clientOrderId': '1656335100',
                    'transactTime': 1656335103251,
                    'price': 20000.0,
                    'origQty': 0.002,
                    'executedQty': 0.0,
                    'cummulativeQuoteQty': 0.0,
                    'status': 'CANCELED',
                    'timeInForce': 'GTC',
                    'type': 'LIMIT',
                    'side': 'BUY',
                    'extraData':
                    {
                        'reduceOnly': False,
                        'stopPrice': 0.0,
                        'workingType': 'CONTRACT_PRICE',
                        'avgPrice': 0.0,
                        'origType': 'LIMIT',
                        'positionSide': 'BOTH',
                        'activatePrice': None,
                        'priceRate': None,
                        'closePosition': False
                    }
                }

        :Notes:

            * Must specify either '**orderId**' or '**localOrderId**'

        """
        return self.exchange.get_order(symbol, orderId, localOrderId, futures=futures)

    def cancelAllSymbolOpenOrders(self, symbol, futures=False):
        """Cancel all active orders of a symbol

        :param symbol: The symbol
        :type symbol: str
        :param futures: False for spot market and True for futures market, defaults to False
        :type futures: bool , optional
        :return: Order datas of the cancelled orders
        :rtype: list(dict)
        :Output:

            .. code-block:: python

                [
                    {
                        'symbol': 'BTCUSDT',
                        'orderId': 3051990476,
                        'clientOrderId': 'sDAclHSkWaO5R3fHYwI9IQ',
                        'transactTime': 1656335835113,
                        'price': 20000.0,
                        'origQty': 0.002,
                        'executedQty': 0.0,
                        'cummulativeQuoteQty': 0.0,
                        'status': 'CANCELED',
                        'timeInForce': 'GTC',
                        'type': 'LIMIT',
                        'side': 'BUY',
                        'extraData':
                        {
                            'reduceOnly': False,
                            'stopPrice': 0.0,
                            'workingType': 'CONTRACT_PRICE',
                            'avgPrice': 0.0,
                            'origType': 'LIMIT',
                            'positionSide': 'BOTH',
                            'activatePrice': None,
                            'priceRate': None,
                            'closePosition': False
                        }
                    },
                    {
                        'symbol': 'BTCUSDT',
                        'orderId': 3051990465,
                        'clientOrderId': 'SxyPikUE9ysMSZ8jbKGQAZ',
                        'transactTime': 1656335835113,
                        'price': 20000.0,
                        'origQty': 0.002,
                        'executedQty': 0.0,
                        'cummulativeQuoteQty': 0.0,
                        'status': 'CANCELED',
                        'timeInForce': 'GTC',
                        'type': 'LIMIT',
                        'side': 'BUY',
                        'extraData':
                        {
                            'reduceOnly': False,
                            'stopPrice': 0.0,
                            'workingType': 'CONTRACT_PRICE',
                            'avgPrice': 0.0,
                            'origType': 'LIMIT',
                            'positionSide': 'BOTH',
                            'activatePrice': None,
                            'priceRate': None,
                            'closePosition': False
                        }
                    }
                ]

        """
        return self.exchange.cancel_all_symbol_open_orders(symbol, futures)

    def cancelOrder(self, symbol, orderId=None, localOrderId=None, futures=False):
        """Cancel an active order

        :param symbol: The Order's symbol
        :type symbol: str
        :param orderId: Exchange assigned '**orderId**' of the order.
        :type orderId: long, optional
        :param localOrderId: User assigned '**clientOrderId**' of the order.
        :type localOrderId: long, optional
        :param futures: False for spot market and True for futures market, defaults to False
        :type futures: bool , optional
        :return: The order data of the canceled order
        :rtype: dict
        :Output:

            .. code-block:: python

                {
                    'symbol': 'BTCUSDT',
                    'orderId': 3051988035,
                    'clientOrderId': '1656335100',
                    'transactTime': 1656335103251,
                    'price': 20000.0,
                    'origQty': 0.002,
                    'executedQty': 0.0,
                    'cummulativeQuoteQty': 0.0,
                    'status': 'CANCELED',
                    'timeInForce': 'GTC',
                    'type': 'LIMIT',
                    'side': 'BUY',
                    'extraData':
                    {
                        'reduceOnly': False,
                        'stopPrice': 0.0,
                        'workingType': 'CONTRACT_PRICE',
                        'avgPrice': 0.0,
                        'origType': 'LIMIT',
                        'positionSide': 'BOTH',
                        'activatePrice': None,
                        'priceRate': None,
                        'closePosition': False
                    }
                }

        :Notes:

            * Must specify either '**orderId**' or '**localOrderId**'

        """
        return self.exchange.cancel_order(symbol, orderId, localOrderId, futures)

    def getTradingFees(self, symbol=None, futures=None):
        """Get the fee structure of the exchange

        :param symbol: Trading fee for a specific symbol
        :type symbol: str, optional
        :param futures: False for spot market and True for futures market, defaults to False
        :type futures: bool , optional
        :return: The fee structure of the exchange
        :rtype: dict
        :Output:

            .. code-block:: python

                {
                    'symbol': 'BTCUSDT',
                    'makerCommission': '0.001',
                    'takerCommission': '0.001'
                }

        :Notes:

            * Specify the symbol for better output.
        """
        return self.exchange.get_trading_fees(symbol=symbol, futures=futures)

    def getSymbolTickerPrice(self, symbol, futures=False):
        """Get the latest price of a symbol

        :param symbol: Symbol
        :type symbol: str
        :param futures: False for spot market and True for futures market, defaults to False
        :type futures: bool , optional
        :return: The latest price of the symbol specified as a float number
        :rtype: float
        :Output:

            .. code-block:: python

                21275.84

        """
        return self.exchange.get_symbol_ticker_price(symbol, futures)

    def getSymbolKlines(
        self,
        symbol,
        interval,
        startTime=None,
        endTime=None,
        limit=None,
        futures=False,
        blvtnav=False,
        convertDateTime=False,
        doClean=False,
        toCleanDataframe=False,
    ):
        """ Get a symbol's Klines (candlestick) data

        :param symbol: The symbol to fetch the klines
        :type symbol: str
        :param interval: The interval of klines data
        :type interval: str
        :param startTime: Timestamp for the start time of the data
        :type startTime: long , optional
        :param endTime: Timestamp for the end time of the data
        :type endTime: long , optional
        :param limit: Number of klines data (candlesticks) to be returned
        :type limit: int , optional
        :param futures: False for spot market and True for futures market, defaults to False
        :type futures: bool , optional
        :param blvtnav: Only for **Binance**. Return the blvtnav data.
        :type blvtnav: bool , optional
        :param convertDateTime: Convert the timestamps to datetime objects in the returned data
        :type convertDateTime: bool , optional
        :param doClean: Only return desired columns of the data
        :type doClean: bool , optional
        :param toCleanDataframe: Returned Pandas dataframe object of the data with desired columns
        :type toCleanDataframe: bool , optional
        :return: Either a 2D array containing the data or a pandas dataframe object of the data
        :rtype: list(list)) or pandas.DataFrame
        :Output as list:

            .. code-block:: python

                [
                    [
                        1656327600000,
                        21437.18,
                        21445.56,
                        21380.02,
                        21428.66,
                        525.2881,
                        1656328499999,
                        11246729.6314567,
                        9626.0, 277.31647,
                        5936870.8288083,
                        0.0
                    ],
                    [
                        1656328500000,
                        21428.66,
                        21450.0,
                        21339.83,
                        21376.49,
                        477.24441,
                        1656329399999,
                        10206890.129678,
                        9503.0, 208.56099,
                        4459849.0304778,
                        0.0
                    ]
                ]

        :Output as DataFrame:

            .. code-block:: python

                                         open      high       low     close   \

                date
                2022-06-27 16:15:00  21367.67  21376.00  21285.48  21304.55
                2022-06-27 16:30:00  21304.56  21329.91  21301.24  21308.73

                                        volume            closeDate  tradesNum
                date
                2022-06-27 16:15:00  470.24040  2022-06-27 16:29:59     8980.0
                2022-06-27 16:30:00  113.05128  2022-06-27 16:44:59     2693.0

        :Notes:

            * '**startTime**' and '**endTime**' variables must be timestamps. If neither are sent, will return the \
            latest data.
            * '**limit**' parameter has a maximum value (usually 1000). if the given limit number is bigger than that, \
            only the maximum number will be fetched.
            * Desired columns are: **open** - **high** - **low** - **close** - **volume** - **closeDate** \
            - **tradesNum**

        """
        return self.exchange.get_symbol_klines(
            symbol,
            interval,
            startTime,
            endTime,
            limit,
            futures,
            blvtnav,
            convertDateTime,
            doClean,
            toCleanDataframe,
        )

    def getExchangeTime(self, futures=False):
        """Get time of the exchange

        :param futures: False for spot market and True for futures market, defaults to False
        :type futures: bool , optional
        :return: Timestamp of the current time on the server
        :rtype: int
        :Output:

            .. code-block:: python

                1656330049381

        """
        return self.exchange.get_exchange_time(futures)

    def createAndTestFuturesOrder(
        self,
        symbol,
        side,
        orderType,
        positionSide=None,
        timeInForce=None,
        quantity=None,
        reduceOnly=None,
        price=None,
        newClientOrderId=None,
        stopPrice=None,
        closePosition=None,
        activationPrice=None,
        callbackRate=None,
        workingType=None,
        priceProtect=None,
        newOrderRespType=None,
        extraParams=None,
        quoteQuantity=None,
    ):
        """ Create a FuturesOrderData object and test the given parameters for validity. The object returned is then used \
        to send an order to the exchange by :func:`make_futures_order() <TradeGate.TradeGate.make_futures_order>`

        :param symbol: Symbol of the order
        :type symbol: str
        :param side: Side of the order. Either '**BUY**' or '**SELL**'
        :type side: str
        :param orderType: Type of the order. can be '**MARKET**', '**LIMIT**' or others (Check exchange's API documentation)
        :type orderType: str
        :param positionSide: Only for binance. '**BOTH**' for One-way Mode and '**LONG**' or '**SHORT**' for Hedge Mode.\
         It must be sent in Hedge Mode.
        :type positionSide: str , optional
        :param timeInForce: Order's time in force. Can be '**GTC**', '**IOC**', '**FOK**' or specific to the exchange. \
        Check the exchange's API documentation for more options.
        :type timeInForce: str , optional
        :param quantity: The amount of base asset of the order.
        :type quantity: float , optional
        :param reduceOnly: A flag to reduce the position size only.
        :type reduceOnly: bool , optional
        :param price: Order price
        :type price: float , optional
        :param newClientOrderId: Custom string to identify your order for yourself.
        :type newClientOrderId: str , optional
        :param stopPrice: Price condition to trigger the order. Only for stop order type.
        :type stopPrice: float , optional
        :param closePosition: Whether to close the current open position or not.
        :type closePosition: bool , optional
        :param activationPrice: Only for Binance. Used with '**TRAILING_STOP_MARKET**' order type.
        :type activationPrice: float , optional
        :param callbackRate: Only for Binance. Used with TRAILING_STOP_MARKET orders and specifies callback percentage.
        :type callbackRate: float , optional
        :param workingType: With what price the stop order is triggered: '**MARK_PRICE**' or '**CONTRACT_PRICE**'
        :type workingType: str , optional
        :param priceProtect: Only for Binance. Whether to protect the stop order from large difference of mark price \
        and contract price.
        :type priceProtect: bool , optional
        :param newOrderRespType: Only for Binance. Response of the order, either '**ACK**' or '**RESULT**'.
        :type newOrderRespType: str , optional
        :param extraParams: Extra parameters for other exchanges than Binance.
        :type extraParams: dict , optional
        :param quoteQuantity: The amount of quote asset of the order.
        :type quoteQuantity: float , optional
        :return: An orderData object of the created order, ready to be submitted using \
        :func:`make_futures_order() <TradeGate.TradeGate.make_futures_order>`
        :rtype: OrderData

        :Notes:

            * Input parameters are based on Binance API. We map the appropriate parameters from input to the exchange \
            requirements, but other parameters must be inside the '**extraParams**' dictionary.
            * Preferably use '**quantity**' instead of '**quoteQuantity**'. Some exchanges don't accept this parameter \
            and we convert it to the quantity based on the price (if provided) or current market price.
            * Some parameters have values that are only valid for some exchanges. See each exchange's documentation for\
             these values.
            * ValueError exception is raised when wrong combination of parameters are sent. Check the exception's \
            message for guidance.
            * Do not send '**price**' with '**MARKET**' orders.
            * You must specify '**price**' for '**LIMIT**' orders.
            * Other than **ByBit**, \
            use :func:`make_sl_tp_limit_futures_order() <TradeGate.TradeGate.make_sl_tp_limit_futures_order>` or \
            :func:`make_sl_tp_market_futures_order() <TradeGate.TradeGate.make_sl_tp_market_futures_order>` to send take profit \
            and stop loss.
            * if '**closePosition**' is used, do not specify '**quantity**' or '**quoteQuantity**'.
            * For **ByBit**:

                * You can send take profit by specifying '**take_profit**' (price) and '**tp_trigger_by**' \
                ('**LastPrice**', '**IndexPrice**' or '**MarkPrice**') in '**extraParams**'.
                * You can send take profit by specifying '**stop_loss**' (price) and '**sl_trigger_by**' \
                ('**LastPrice**', '**IndexPrice**' or '**MarkPrice**') in '**extraParams**'.

            * For **KuCoin**:

                * You must send '**leverage**' parameter inside '**extraParams**' for the **KuCoin** exchange.
                * Send '**stop**' (either '**down**' or '**up**') and '**stopPriceType**' (either '**TP**', '**IP**' \
                or '**MP**') in the '**extraParams**' when specifying '**stopPrice**'.
                * You can send '**postOnly**', '**hidden**', '**iceberg**' and '**visibleSize**' parameters inside \
                '**extraParams**'.
                * There are two values for the '**timeInForce**' variable: '**GTC**' (Good Till Cancel) and '**IOC**' \
                (Immediate Or Cancel)


        """
        return self.exchange.create_and_test_futures_order(
            symbol,
            side,
            orderType,
            positionSide,
            timeInForce,
            quantity,
            reduceOnly,
            price,
            newClientOrderId,
            stopPrice,
            closePosition,
            activationPrice,
            callbackRate,
            workingType,
            priceProtect,
            newOrderRespType,
            extraParams,
        )

    def makeFuturesOrder(self, futuresOrderData):
        """Make a futures order

        :param futuresOrderData: OrderData created using :func:`make_futures_order() <TradeGate.TradeGate.make_futures_order>`
        :type futuresOrderData: OrderData
        :return: submitted order information
        :rtype: dict
        :Output:

            .. code-block:: python

                {
                    'symbol': 'BTCUSDT',
                    'orderId': 3049327900,
                    'clientOrderId': 'aLHt4lEJzO2Xeh21GebGCz0',
                    'transactTime': 1655704960910,
                    'price': 0.0,
                    'origQty': 0.003,
                    'executedQty': 0.0,
                    'cummulativeQuoteQty': 0.0,
                    'status': 'NEW',
                    'timeInForce': 'GTC',
                    'type': 'MARKET',
                    'side': 'BUY',
                    'extraData':
                        {
                            'reduceOnly': False,
                            'stopPrice': 0.0,
                            'workingType':
                            'CONTRACT_PRICE',
                            'avgPrice': 0.0,
                            'origType': 'MARKET',
                            'positionSide': 'BOTH',
                            'activatePrice': None,
                            'priceRate': None,
                            'closePosition': False
                        }
                    }

        """
        return self.exchange.make_futures_order(futuresOrderData)

    def makeBatchFuturesOrder(self, batchOrders):
        """ Make multiple futures order with single request

        :param batchOrders: list of OrderDatas created using :func:`make_futures_order() <TradeGate.TradeGate.make_futures_order>`
        :type batchOrders: list(OrderData)
        :return: List of order information submitted
        :rtype: list(dict)
        :Output:

            .. code-block:: python

                [
                    {
                        'symbol': 'BTCUSDT',
                        'orderId': 3049327900,
                        'clientOrderId': 'aLHt4lEJzO2Xeh21GebGCz0',
                        'transactTime': 1655704960910,
                        'price': 0.0,
                        'origQty': 0.003,
                        'executedQty': 0.0,
                        'cummulativeQuoteQty': 0.0,
                        'status': 'NEW',
                        'timeInForce': 'GTC',
                        'type': 'MARKET',
                        'side': 'BUY',
                        'extraData':
                        {
                            'reduceOnly': False,
                            'stopPrice': 0.0,
                            'workingType':
                            'CONTRACT_PRICE',
                            'avgPrice': 0.0,
                            'origType': 'MARKET',
                            'positionSide': 'BOTH',
                            'activatePrice': None,
                            'priceRate': None,
                            'closePosition': False
                        }
                    },
                    {
                        'symbol': 'BTCUSDT',
                        'orderId': 3049327901,
                        'clientOrderId': 'hUacgz6xW3Vp71sc8yztxK1',
                        'transactTime': 1655704960910,
                        'price': 0.0,
                        'origQty': 0.005,
                        'executedQty': 0.0,
                        'cummulativeQuoteQty': 0.0,
                        'status': 'NEW',
                        'timeInForce': 'GTC',
                        'type': 'MARKET',
                        'side': 'BUY',
                        'extraData':
                        {
                            'reduceOnly': False,
                            'stopPrice': 0.0,
                            'workingType':
                            'CONTRACT_PRICE',
                            'avgPrice': 0.0,
                            'origType': 'MARKET',
                            'positionSide': 'BOTH',
                            'activatePrice': None,
                            'priceRate': None,
                            'closePosition': False
                        }
                    }
                ]

        :Notes:

            * Not available for **KuCoin** exchange.
            * The execution of orders in the batch are not dependent on each other, so be careful if orders depend on \
            one another.

        """
        return self.exchange.make_batch_futures_order(batchOrders)

    def changeInitialLeverage(self, symbol, leverage):
        """Change initial leverage for a symbol

        :param symbol: Futures symbol
        :type symbol: str
        :param leverage: Initial leverage number.
        :type leverage: int
        :return: The number of leverage
        :rtype: int
        :Output:

            .. code-block:: python

                10

        :Notes:

            * Not available for **KuCoin** exchange. You must specify leverage when sending an order for this exchange.

        """
        return self.exchange.change_initial_leverage(symbol, leverage)

    def changeMarginType(self, symbol, marginType, params=None):
        """ Change position margin of a symbol

        :param symbol: Symbol of the position
        :type symbol: str
        :param marginType: Type of the margin. Either '**ISOLATED**' or '**CROSSED**'.
        :type marginType: str
        :param params: Extra data needed for some exchanges.
        :type params: dict , optional
        :return: True if changing was successful or False if unsuccessful
        :rtype: bool
        :Output:

            .. code-block:: python

                True

        :Notes:

            * For **KuCoin**, '**CROSSED**' means enabling '**auto_add_margin**', and '**ISOLATED**' means disabling it.
            * For **ByBit**, you must specify '**buyLeverage**' and '**sellLeverage**' inside params. If switching from \
            '**CROSSED**' to '**ISOLATED**', theses two numbers must be equal.

        """
        return self.exchange.change_margin_type(symbol, marginType, params)

    def changePositionMargin(self, symbol, amount):
        """ Change position margin of a symbol

        :param symbol: Symbol of the position
        :type symbol: str
        :param amount: Amount to be added (or subtracted)
        :type amount: float
        :return: True if changing was successful or False if unsuccessful
        :rtype: bool
        :Output:

            .. code-block:: python

                True

        :Notes:

            * The amount can be positive or negative for **Binance** exchange but it must be positive for other exchanges.
            * Use :func:`change_margin_type() <TradeGate.TradeGate.change_margin_type>` to make current position available \
            for changing amount.

        """
        return self.exchange.change_position_margin(symbol, amount)

    def spotBestBidAsks(self, symbol=None):
        """Returns best bid and best ask price with their quantities

        :param symbol: Symbol name of the orders
        :type symbol: str
        :return: A dictionary with best bid and ask with their quantity
        :rtype: dict
        :Output:

            .. code-block:: python

                {
                    'symbol': 'BTC-USDT',
                    'bidPrice': '18125.7',
                    'bidQty': '0.00839998',
                    'askPrice': '18126',
                    'askQty': '0.00496777'
                }

        """
        return self.exchange.spot_best_bid_asks(symbol)

    def getSymbolOrderBook(self, symbol, limit=None, futures=False):
        """Returns list of current orders in the orderbook of the exchange

        :param symbol: Symbol name of the orders
        :type symbol: str
        :param limit: Maximum number of returned bids and asks
        :type limit: int , Optional
        :param futures: False for spot market and True for futures market, defaults to False
        :type futures: bool , optional
        :return: A dictionary with bids and asks. Each bid and ask is a tuple of price and quantity.
        :rtype: dict
        :Output (limit=5):

            .. code-block:: python

                {
                    'lastUpdateId': 1630819351623,
                    'bids': [
                        ['18009.90', '0.557'],
                        ['18009.80', '0.076'],
                        ['18009.30', '0.002'],
                        ['18009.20', '0.004'],
                        ['18008.20', '0.038']
                    ],
                    'asks': [
                        ['18010.00', '0.464'],
                        ['18010.50', '0.101'],
                        ['18010.70', '0.099'],
                        ['18010.90', '0.017'],
                        ['18011.20', '0.206']
                    ]
                }

        """
        return self.exchange.get_symbol_order_book(symbol, limit, futures)

    def getSymbolRecentTrades(self, symbol, limit=None, futures=False):
        """Returns list of the recent trades for a symbol

        :param symbol: Symbol name of the trades
        :type symbol: str
        :param limit: Maximum number of returned trade datas
        :type limit: int , Optional
        :param futures: False for spot market and True for futures market, defaults to False
        :type futures: bool , optional
        :return: A data frame of the trade infos
        :rtype: pandas.DataFrame
        :Output columns:

            * **id** (:py:class:`int`) - ID of the trade
            * **price** (:py:class:`float`) - Price of the trade
            * **qty** (:py:class:`float`) - Amount of the base asset
            * **quoteQty** (:py:class:`float`) - Amount of the quote asset
            * **time** (:py:class:`int`) - Timestamp of the trade
            * **isBuyerMaker** (:py:class:`bool`) - If trade is buyer-maker

        """
        return self.exchange.get_symbol_recent_trades(symbol, limit, futures)

    def symbolAccountTradeHistory(self, symbol, fromId=None, limit=None, futures=False):
        """Returns list of the trade history for user orders

        :param symbol: Symbol name of the trades
        :type symbol: str
        :param fromId: Only return trades from the specified id forward.
        :type fromId: str , Optional
        :param limit: Maximum number of returned trade datas
        :type limit: int , Optional
        :param futures: False for spot market and True for futures market, defaults to False
        :type futures: bool , optional
        :return: A list of trade datas
        :rtype: list(dict)
        :Output:

            .. code-block:: python

                [
                    {
                        'symbol': 'BTCUSDT',
                        'id': 233453846,
                        'orderId': 3046255912,
                        'orderListId': None,
                        'price': 27426.7,
                        'qty': 0.61,
                        'quoteQty': 16730.287,
                        'commission': 0.0,
                        'commissionAsset': 'USDT',
                        'time': 1655001759033,
                        'isBuyer': False,
                        'isMaker': False,
                        'isBestMatch': None
                    },
                    {
                        'symbol': 'BTCUSDT',
                        'id': 233480351,
                        'orderId': 3046360099,
                        'orderListId': None,
                        'price': 27300.0,
                        'qty': 0.002,
                        'quoteQty': 54.6,
                        'commission': 0.02184,
                        'commissionAsset': 'USDT',
                        'time': 1655027380068,
                        'isBuyer': True,
                        'isMaker': False,
                        'isBestMatch': None
                    },
                    ...
                ]

        :Notes:

            * The **orderListId** returned parameter is either None or -1 if order was made with API.
            * The **isBestMatch** returned parameter is not reliable and not available on most of the exchanges.
        """
        return self.exchange.symbol_account_trade_history(
            symbol=symbol, futures=futures, from_id=fromId, limit=limit
        )

    def makeSlTpLimitFuturesOrder(
        self,
        symbol,
        orderSide,
        enterPrice,
        quantity=None,
        quoteQuantity=None,
        takeProfit=None,
        stopLoss=None,
        leverage=None,
        marginType=None,
    ):
        """Make market price futures order with take profit and stop loss.

        :param symbol: Name of the symbol
        :type symbol: str
        :param orderSide: Side of the order.
        :type orderSide: str
        :param enterPrice: Limit price of the order
        :type enterPrice: float , optional
        :param quantity: Amount of the base asset
        :type quantity: float , optional
        :param quoteQuantity: Amount of the quote asset
        :type quoteQuantity: float , optional
        :param takeProfit: Take profit price
        :type takeProfit: float , optional
        :param stopLoss: Stop loss price
        :type stopLoss: float , optional
        :param leverage: Desired leverage for the order.
        :type leverage: int , optional
        :param marginType: Margin type of the order
        :type marginType: str , optional
        :return: Returns orderID's of the orders submitted.
        :rtype: dict
        :Output:

            .. code-block:: python

                {
                    'mainOrder': 3048448085,
                    'stopLoss': 3048448086,
                    'takeProfit': 3048448087
                }

        :Notes:

            * The **orderSide** parameter can either be **BUY** or **SELL**.
            * Should specify either **quantity** or **quoteQuantity**.
            * The **leverage** parameter is mandatory for the following exchanges:

                * KuCoin

            * The **marginType** is currently only valid for **Binance** exchange.
            * Be careful with the **takeProfit** and **stopLoss** prices, if they would trigger immidietly, there will be an error.
            * Use with a try catch block preferably.

        """
        return self.exchange.make_sl_tp_limit_futures_order(
            symbol,
            orderSide,
            quantity,
            quoteQuantity,
            enterPrice,
            takeProfit,
            stopLoss,
            leverage,
            marginType,
        )

    def makeSlTpMarketFuturesOrder(
        self,
        symbol,
        orderSide,
        quantity=None,
        quoteQuantity=None,
        takeProfit=None,
        stopLoss=None,
        leverage=None,
        marginType=None,
    ):
        """Make market price futures order with take profit and stop loss.

        :param symbol: Name of the symbol
        :type symbol: str
        :param orderSide: Side of the order.
        :type orderSide: str
        :param quantity: Amount of the base asset
        :type quantity: float , optional
        :param quoteQuantity: Amount of the quote asset
        :type quoteQuantity: float , optional
        :param takeProfit: Take profit price
        :type takeProfit: float , optional
        :param stopLoss: Stop loss price
        :type stopLoss: float , optional
        :param leverage: Desired leverage for the order.
        :type leverage: int , optional
        :param marginType: Margin type of the order
        :type marginType: str , optional
        :return: Returns orderID's of the orders submitted.
        :rtype: dict
        :Output:

            .. code-block:: python

                {
                    'mainOrder': 3048424829,
                    'stopLoss': 3048424830,
                    'takeProfit': 3048424828
                }

        :Notes:

            * The **orderSide** parameter can either be **BUY** or **SELL**.
            * Should specify either **quantity** or **quoteQuantity**.
            * The **leverage** parameter is mandatory for the following exchanges:

                * KuCoin
            * The **marginType** is currently only valid for **Binance** exchange.
            * Be careful with the **takeProfit** and **stopLoss** prices, if they would trigger immidietly, there will be an error.
            * Use with a try catch block preferably.

        """
        return self.exchange.make_sl_tp_market_futures_order(
            symbol,
            orderSide,
            quantity,
            quoteQuantity,
            takeProfit,
            stopLoss,
            leverage,
            marginType,
        )

    def getPositionInfo(self, symbol=None):
        """Returns information of  position or positions. (Only for futures accounts)

        :param symbol: The symbol of the position
        :type symbol: str , optional
        :return: A list of position information. If the symbol parameter is given, the list will contain only one element.
        :rtype: list(dict)
        :Output with symbol specified:

            .. code-block:: python

                [
                    {
                        'entryPrice': 19229.6,
                        'isAutoAddMargin': True,
                        'leverage': 10.0,
                        'maxNotionalValue': 1000000.0,
                        'liquidationPrice': 20935.5130297,
                        'markPrice': 19222.55166016,
                        'positionAmt': -0.01,
                        'symbol': 'BTCUSDT',
                        'unrealizedProfit': 0.07048339,
                        'marginType': 'isolated',
                        'isolatedMargin': 19.22316499,
                        'positionSide': 'BOTH'
                    }
                ]

        :Output without symbol specified:

            .. code-block:: python

                [
                    {
                        'entryPrice': 0.0,
                        'isAutoAddMargin': True,
                        'leverage': 20.0,
                        'maxNotionalValue': 25000.0,
                        'liquidationPrice': 0.0,
                        'markPrice': 0.0,
                        'positionAmt': 0.0,
                        'symbol': 'RAYUSDT',
                        'unrealizedProfit': 0.0,
                        'marginType': 'cross',
                        'isolatedMargin': 0.0,
                        'positionSide': 'BOTH'
                    },
                    {
                        'entryPrice': 0.0,
                        'isAutoAddMargin': True,
                        'leverage': 20.0,
                        'maxNotionalValue': 25000.0,
                        'liquidationPrice': 0.0,
                        'markPrice': 0.0,
                        'positionAmt': 0.0,
                        'symbol': 'API3USDT',
                        'unrealizedProfit': 0.0,
                        'marginType': 'cross',
                        'isolatedMargin': 0.0,
                        'positionSide': 'BOTH'
                    },
                    ...
                ]

        """
        return self.exchange.get_position_info(symbol)

    def getSymbolMinTrade(self, symbol, futures=False):
        """Returns information of valid minimum quantity, quote quantity and price precision.

        :param symbol: The symbol which the information is for.
        :type symbol: str
        :param futures: False for spot market and True for futures market, defaults to False
        :type futures: bool , optional
        :return: A dictionary containing information about the minimum valid values of the specified symbol
        :rtype: dict
        :Output parameters:

            .. list-table::
               :widths: 10 50
               :header-rows: 0

               * - **stepPrice**
                 - Price's maximum precision
               * - **minQuantity**
                 - Minimum valid quantity
               * - **precisionStep**
                 - Quantity's maximum precision
               * - **minQuoteQuantity**
                 - Minimum valid quote quantity

        :Output:

            .. code-block:: python

                {
                    'stepPrice': 0.01,
                    'minQuantity': 1e-05,
                    'precisionStep': 1e-05,
                    'minQuoteQuantity': 0.19279200000000002
                }

        """
        return self.exchange.get_symbol_min_trade(symbol, futures)

    def getIncomeHistory(
        self, symbol, incomeType=None, startTime=None, endTime=None, limit=None
    ):
        """Returns list of changes to the account balance. (Only for futures accounts)

        :param symbol: The symbol which the incomes are related to
        :type symbol: str
        :param incomeType: Type of income. For options visit the exchange's API documentation.
        :type incomeType: str , Optional
        :param startTime: If specified, incomes will be fetched from that time forward. (Time string format : '%Y-%m-%d %H:%M:%S')
        :type startTime: str , Optional
        :param endTime: If specified, incomes will be fetched until that time. (Time string format : '%Y-%m-%d %H:%M:%S')
        :type endTime: str , Optional
        :param limit: Maximum number of entries returned. For available options visit the exchange's API documentation.
        :type limit: int
        :return: A list of incomes
        :rtype: list(dict)
        :Output:

            .. code-block:: python

                [
                    {
                        'symbol': 'BTCUSDT',
                        'incomeType': 'FUNDING_FEE',
                        'income': 0.26403463,
                        'asset': 'USDT',
                        'time': 1655280000000
                    },
                    {
                        'symbol': 'BTCUSDT',
                        'incomeType': 'REALIZED_PNL',
                        'income': 26.0335,
                        'asset': 'USDT',
                        'time': 1655281638000
                    },
                    {
                        'symbol': 'BTCUSDT',
                        'incomeType': 'COMMISSION',
                        'income': -0.08219559,
                        'asset': 'USDT',
                        'time': 1655281638000
                    },
                    ...
                ]

        """

        return self.exchange.get_income_history(
            symbol, incomeType, startTime, endTime, limit
        )

    def getSymbolList(self, futures=False):
        """Returns list of symbol names available for trade

        :param futures: False for spot market and True for futures market, defaults to False
        :type futures: bool , optional
        :return: Returns a list of strings
        :rtype: list(str)
        :Output:

            .. code-block:: python

                [
                    'BTCUSDT',
                    'ETHUSDT',
                    'BCHUSDT',
                    'XRPUSDT',
                    'EOSUSDT',
                    'LTCUSDT',
                    'TRXUSDT',
                    ...
                ]

        """
        return self.exchange.get_symbol_list(futures=futures)

    def getSymbol24hChanges(self, futures=False):
        """Returns all symbols 24-hour change percentages

        :param futures: False for spot market and True for futures market, defaults to False
        :type futures: bool , optional
        :return: Returns a list of tuples containing asset names and percentage of change in 24-hour
        :rtype: list(tuple)
        :Output:

            .. code-block:: python

                [
                    ('PONDUSDT', 28.45),
                    ('PONDBTC', 28.261),
                    ('PONDBUSD', 28.162),
                    ('NULSBTC', 24.321),
                    ('NULSUSDT', 23.975),
                    ('NULSBUSD', 23.244),
                    ('CTXCBTC', 20.551),
                    ('CTXCUSDT', 19.959),
                    ('CTXCBUSD', 19.776),
                    ...
                ]

        """
        return self.exchange.get_symbol_24h_changes(futures=futures)

    def getLatestSymbolNames(self, numOfSymbols=None, futures=True):
        """Returns list of newly added symbols to the exchange. Currently, only working for futures market.

        :param numOfSymbols: Number of symbols returned, sorted for the newest to oldest.
        :type numOfSymbols: int, optional
        :param futures: False for spot market and True for futures market, defaults to False
        :type futures: bool , optional
        :return: Returns a list of tuples containing asset names and a datetime object specifying its listed date.
        :rtype: list(tuple)
        :Output:

            .. code-block:: python

                [
                    ('DOTBUSD', datetime.datetime(2022, 6, 7, 11, 30)),
                    ('TLMBUSD', datetime.datetime(2022, 6, 7, 11, 30)),
                    ('ICPBUSD', datetime.datetime(2022, 6, 7, 11, 30)),
                    ('OPUSDT', datetime.datetime(2022, 6, 1, 11, 30)),
                    ('LUNA2BUSD', datetime.datetime(2022, 5, 31, 11, 30)),
                    ('1000LUNCBUSD', datetime.datetime(2022, 5, 30, 11, 30)),
                    ('GALABUSD', datetime.datetime(2022, 5, 25, 11, 30)),
                    ('TRXBUSD', datetime.datetime(2022, 5, 25, 11, 30)),
                    ('DODOBUSD', datetime.datetime(2022, 5, 24, 11, 30)),
                    ('ANCBUSD', datetime.datetime(2022, 5, 24, 11, 30))
                ]

        """
        return self.exchange.get_latest_symbol_names(
            num_of_symbols=numOfSymbols, futures=futures
        )

    def getLongShortRatios(
        self, symbol, period, limit=None, startTime=None, endTime=None
    ):
        print(self.exchange)
        return self.exchange.get_long_short_ratios(
            symbol, period, limit, startTime, endTime
        )

    def getDepositAddress(self, coin, network=None):
        return self.exchange.get_deposit_address(coin, network)

    def withdraw(self, coin, address, amount, extra_data):
        return self.exchange.withdraw(
            coin=coin, amount=amount, address=address, **extra_data
        )



import json
from binance_f.exception.binanceapiexception import BinanceApiException

class JsonWrapper:
    def __init__(self, json_object):
        self.json_object = json_object

    def __check_mandatory_field(self, name):
        if name not in self.json_object:
            raise BinanceApiException(BinanceApiException.RUNTIME_ERROR,
                                    "[Json] Get json item field: " + name + " does not exist")

    def contain_key(self, name):
        if name in self.json_object:
            return True
        else:
            return False

    def get_boolean(self, name):
        self.__check_mandatory_field(name)
        return bool(self.json_object[name])

    def get_string(self, name):
        self.__check_mandatory_field(name)
        return str(self.json_object[name])

    def get_int(self, name):
        self.__check_mandatory_field(name)
        return int(self.json_object[name])

    def get_string_or_default(self, name, default):
        if self.contain_key(name):
            return str(self.json_object[name])
        else:
            return default

    def get_int_or_default(self, name, default):
        if self.contain_key(name):
            return int(self.json_object[name])
        else:
            return default

    def get_float(self, name):
        self.__check_mandatory_field(name)
        return float(self.json_object[name])

    def get_float_or_default(self, name, default):
        if self.contain_key(name):
            return float(self.json_object[name])
        else:
            return default

    def get_object(self, name):
        self.__check_mandatory_field(name)
        return JsonWrapper(self.json_object[name])

    def get_object_or_default(self, name, defalut_value):
        if name not in self.json_object:
            return defalut_value
        else:
            return JsonWrapper(self.json_object[name])

    def get_array(self, name):
        self.__check_mandatory_field(name)
        return JsonWrapperArray(self.json_object[name])

    def convert_2_array(self):
        return JsonWrapperArray(self.json_object)

    def convert_2_dict(self):
        items = dict()
        for item in self.json_object:
            name = item
            items[name] = self.json_object[name]
        return items

    def convert_2_list(self):
        items = list()
        for item in self.json_object:
            items.append(item)
        return items



class JsonWrapperArray:
    def __init__(self, json_object):
        self.json_object = json_object

    def get_items(self):
        items = list()
        for item in self.json_object:
            items.append(JsonWrapper(item))
        return items

    def get_items_as_array(self):
        items = list()
        for item in self.json_object:
            items.append(JsonWrapperArray(item))
        return items

    def get_float_at(self, index):
        return float(self.json_object[index])

    def get_items_as_string(self):
        items = list()
        for item in self.json_object:
            items.append(str(item))
        return items

    def get_array_at(self, index):
        return JsonWrapperArray(self.json_object[index])

    def get_object_at(self, index):
        return JsonWrapper(self.json_object[index])





from kucoin_futures.base_request.base_request import KucoinFuturesBaseRestApi


class MarketData(KucoinFuturesBaseRestApi):

    def get_server_timestamp(self):
        """
        https://docs.kumex.com/#server-time

        get server timestamp
        :return: 1570609496404
        """
        return self._request('GET', '/api/v1/timestamp', auth=False)

    def get_interest_rate(self, symbol, startAt=None, endAt=None, reverse=True, offset=0, forward=True, maxCount=10):
        """
        https://docs.kumex.com/#get-interest-rate-list

        :param symbol: interest symbol (Mandatory)
        :type: str
        :param startAt: start time(milisecond) (optional)
        :type: int
        :param endAt:  end time(milisecond) (optional)
        :type: int
        :param reverse: is reverse? (optional)
        :type: bool
        :param offset:  Start offset. The unique attribute of the last returned result of the last request. The data of (optional)
        the first page will be returned by default
        :type: int
        :param forward: his parameter functions to judge whether the lookup is forward or not (optional)
        :type: bool
        :param maxCount:Max record count (optional)
        :type : int
        :return:
            {'dataList':
                [
                    {'symbol': '.XBTINT', 'granularity': 60000, 'timePoint': 1570611840000, 'value': 0.0003},
                    {'symbol': '.XBTINT', 'granularity': 60000, 'timePoint': 1570611780000, 'value': 0.0003},
                    {'symbol': '.XBTINT', 'granularity': 60000, 'timePoint': 1570611720000, 'value': 0.0003},
                    {'symbol': '.XBTINT', 'granularity': 60000, 'timePoint': 1570611660000, 'value': 0.0003},
                    {'symbol': '.XBTINT', 'granularity': 60000, 'timePoint': 1570611600000, 'value': 0.0003},
                     {'symbol': '.XBTINT', 'granularity': 60000, 'timePoint': 1570611540000, 'value': 0.0003},
                     {'symbol': '.XBTINT', 'granularity': 60000, 'timePoint': 1570611480000, 'value': 0.0003},
                      {'symbol': '.XBTINT', 'granularity': 60000, 'timePoint': 1570611420000, 'value': 0.0003},
                      {'symbol': '.XBTINT', 'granularity': 60000, 'timePoint': 1570611360000, 'value': 0.0003},
                      {'symbol': '.XBTINT', 'granularity': 60000, 'timePoint': 1570611300000, 'value': 0.0003}
                        ],
                    'hasMore': True}
        """

        params = {'symbol': symbol}

        if startAt:
            params['startAt'] = startAt
        if endAt:
            params['endAt'] = endAt
        if reverse:
            params['reverse'] = reverse
        if offset:
            params['offset'] = offset
        if forward:
            params['forward'] = forward
        if maxCount:
            params['maxCount'] = maxCount

        return self._request('GET', '/api/v1/interest/query', auth=False, params=params)

    def get_index_list(self, symbol, startAt=None, endAt=None, reverse=True, offset=0, forward=True, maxCount=10):
        """
        https://docs.kumex.com/#get-index-list

        :param symbol: interest symbol (Mandatory)
        :type: str
        :param startAt: start time(milisecond) (optional)
        :type: int
        :param endAt:  end time(milisecond) (optional)
        :type: int
        :param reverse: is reverse? (optional)
        :type: bool
        :param offset:  Start offset. The unique attribute of the last returned result of the last request. The data of (optional)
        the first page will be returned by default
        :type: int
        :param forward: his parameter functions to judge whether the lookup is forward or not (optional)
        :type: bool
        :param maxCount:Max record count (optional)
        :type : int
        :return:
            {'dataList':
                [
                    {'symbol': '.BXBT','decomposionList': [
                            {'price': 8214.62, 'weight': 0.05746732, 'exchange': 'gemini'},
                            {'price': 8212.4, 'weight': 0.1896515, 'exchange': 'kraken'},
                            {'price': 8206.35, 'weight': 0.43039379, 'exchange': 'coinbase'},
                            {'price': 8221.60243, 'weight': 0.03994456, 'exchange': 'liquid'},
                            {'price': 8211.981, 'weight': 0.02333998, 'exchange': 'bittrex'}, {
                            'price': 8206.47, 'weight': 0.25920285, 'exchange': 'bitstamp'}],
                            'granularity': 5000, 'timePoint': 1570612465000, 'value': 8208.74},
                    {'symbol': '.BXBT','decomposionList': [
                            {'price': 8214.62, 'weight': 0.05746732, 'exchange': 'gemini'},
                            {'price': 8212.4, 'weight': 0.1896515, 'exchange': 'kraken'},
                            {'price': 8208.98, 'weight': 0.43039379, 'exchange': 'coinbase'},
                            {'price': 8221.60243, 'weight': 0.03994456, 'exchange': 'liquid'},
                            {'price': 8211.981, 'weight': 0.02333998, 'exchange': 'bittrex'},
                            {'price': 8207.47, 'weight': 0.25920285, 'exchange': 'bitstamp'}],
                            'granularity': 5000, 'timePoint': 1570612460000, 'value': 8210.14},
                    {......
                ],
            'hasMore': True}
        """
        params = {'symbol': symbol}

        if startAt:
            params['startAt'] = startAt
        if endAt:
            params['endAt'] = endAt
        if reverse:
            params['reverse'] = reverse
        if offset:
            params['offset'] = offset
        if forward:
            params['forward'] = forward
        if maxCount:
            params['maxCount'] = maxCount

        return self._request('GET', '/api/v1/index/query', auth=False, params=params)

    def get_current_mark_price(self, symbol):
        """
        https://docs.kumex.com/#get-current-mark-price

        :param symbol:
        :type: str
        :return: {'symbol': 'XBTUSDM', 'indexPrice': 8194.22, 'granularity': 5000, 'timePoint': 1570613025000, 'value': 8194.49}
        """

        return self._request('GET', '/api/v1/mark-price/{symbol}/current'.format(symbol=symbol), auth=False)

    def get_premium_index(self, symbol, startAt=None, endAt=None, reverse=True, offset=0, forward=True, maxCount=10):
        """
        https://docs.kumex.com/#get-premium-index

        :param symbol: interest symbol (Mandatory)
        :type: str
        :param startAt: start time(milisecond) (optional)
        :type: int
        :param endAt:  end time(milisecond) (optional)
        :type: int
        :param reverse: is reverse? (optional)
        :type: bool
        :param offset:  Start offset. The unique attribute of the last returned result of the last request. The data of (optional)
        the first page will be returned by default
        :type: int
        :param forward: his parameter functions to judge whether the lookup is forward or not (optional)
        :type: bool
        :param maxCount:Max record count (optional)
        :type : int
        :return:
         {
            "dataList": [
                      {
                        "symbol": ".XBTUSDMPI",              //Premium index symbol
                        "granularity": 60000,                //Granularity (milisecond)
                        "timePoint": 1558000320000,          //Time point (milisecond)
                        "value": 0.022585                    //Premium index
                      },
                      {
                        "symbol": ".XBTUSDMPI",
                        "granularity": 60000,
                        "timePoint": 1558000260000,
                        "value": 0.022611
                      },
                    ......
                ],
            "hasMore": true                        //Whether there are more pages
        }
        """

        params = {'symbol': symbol}

        if startAt:
            params['startAt'] = startAt
        if endAt:
            params['endAt'] = endAt
        if reverse:
            params['reverse'] = reverse
        if offset:
            params['offset'] = offset
        if forward:
            params['forward'] = forward
        if maxCount:
            params['maxCount'] = maxCount

        return self._request('GET', '/api/v1/premium/query', auth=False, params=params)

    def get_current_fund_rate(self, symbol):
        """
        https://docs.kumex.com/#get-current-funding-rate

        :param symbol:  type str (Mandatory)
        :return:
          {
                "symbol": ".XBTUSDMFPI8H",              //Funding Rate Symbol
                "granularity": 28800000,               //Granularity (milisecond)
                "timePoint": 1558000800000,            //Time point (milisecond)
                "value": 0.00375,                      //Funding rate
                "predictedValue": 0.00375              //Predicted funding rate
            }

        """

        return self._request('GET', '/api/v1/funding-rate/{symbol}/current'.format(symbol=symbol), auth=False)

    def get_trade_history(self, symbol):
        """
        https://docs.kumex.com/#transaction-history

        :param symbol:  type str (Mandatory)
        :return:
            [{
            "sequence": 102,                            //Sequence number
            "tradeId": "5cbd7377a6ffab0c7ba98b26",      //Transaction ID
            "takerOrderId": "5cbd7377a6ffab0c7ba98b27", //Taker order ID
            "makerOrderId": "5cbd7377a6ffab0c7ba98b28", //Maker order ID
            "price": "7000.0",                          //Filled price
            "size": 0.1,                                //Filled quantity
            "side": "buy",                              //Side-taker
            "ts": 1545904567062140823                   //Filled time - nanosecond
        },
        .......]

        """
        params = {
            'symbol': symbol
        }

        return self._request('GET', '/api/v1/trade/history', auth=False, params=params)

    def l2_order_book(self, symbol):
        """
        https://docs.kumex.com/#get-full-order-book-level-2

        :param symbol: type tar (Mandatory)
        :return:
        {
              "symbol": "XBTUSDM",      //Symbol
              "sequence": 100,          //Ticker sequence number
              "asks": [
                ["5000.0", 1000],       //Price, quantity
                ["6000.0", 1983],       //Price, quantity
                ......
              ],
              "bids": [
                ["3200.0", 800],        //Price, quantity
                ["3100.0", 100],        //Price, quantity
                ......
              ]
            }
        """

        params = {
            "symbol": symbol
        }
        return self._request('GET', '/api/v1/level2/snapshot', auth=False, params=params)

    def l2_part_order_book(self, symbol, depth=20):
        """
        https://docs.kucoin.com/futures/#get-part-order-book-level-2
        :param symbol: type tar (Mandatory)
        :return:
            {
            "code": "200000",
            "data": {
              "symbol": "XBTUSDM",      //Symbol
              "sequence": 100,          //Ticker sequence number
              "asks": [
                ["5000.0", 1000],   //Price, quantity
                ["6000.0", 1983]        //Price, quantity
              ],
              "bids": [
                ["3200.0", 800],        //Price, quantity
                ["3100.0", 100]     //Price, quantity
              ]
            }
            }
        """

        params = {
            "symbol": symbol
        }
        return self._request('GET', f'/api/v1/level2/depth{depth}', auth=False, params=params)

    def get_l2_messages(self, symbol, start, end):
        """

        :param symbol: type tar (Mandatory)
        :type: str
        :param start: Start sequence number (included in the returned data) (Mandatory)
        :type: int
        :param end:	End sequence number (included in the returned data) (Mandatory)
        :type: int
        :return:
        """
        params = {
            'symbol': symbol,
            'start': start,
            'end': end
        }

        return self._request('GET', '/api/v1/level2/message/query', auth=False, params=params)

    def l3_order_book(self, symbol):
        """
        https://docs.kumex.com/#get-full-order-book-level-3

        :param symbol: type tar (Mandatory)
        :return:
          {
            "code": "200000",
            "data": {
                "symbol": "XBTUSDM",        //Symbol
              "sequence":  100,     //The sequence number of the last received message in building a Level 3 order book
              "bids": [[5567483701231, "dfa123124", "123.12312", 10, 5567483701231], ...],  //Selling data: order placing time - nanosecond, order ID, price, quantity, time at which the order enters the order book -  nanosecond
              "asks": [[5567483701231, "dfa123124", "123.12312", 10, 5567483701231], ...]   //Buying data: order placing time - nanosecond, order ID, price, quantity, time at which the order enters the order book- nanosecond
            }
          }
        """

        params = {
            "symbol": symbol
        }
        return self._request('GET', '/api/v1/level3/snapshot', auth=False, params=params)

    def l3_order_book_v2(self, symbol):
        """
        https://docs.kucoin.com/futures/#get-full-order-book-level-3-v2
        :param symbol: type tar (Mandatory)
        :return:
          {
            "code": "200000",
            "data": {
                "symbol": "XBTUSDM",        //Symbol
              "sequence":  100,     //The sequence number of the last received message in building a Level 3 order book
              "bids": [[5567483701231, "dfa123124", "123.12312", 10, 5567483701231], ...],  //Selling data: order placing time - nanosecond, order ID, price, quantity, time at which the order enters the order book -  nanosecond
              "asks": [[5567483701231, "dfa123124", "123.12312", 10, 5567483701231], ...]   //Buying data: order placing time - nanosecond, order ID, price, quantity, time at which the order enters the order book- nanosecond
            }
          }
        """

        params = {
            "symbol": symbol
        }
        return self._request('GET', '/api/v2/level3/snapshot', auth=False, params=params)

    def get_l3_messages(self, symbol, start, end):
        """

        :param symbol: type tar (Mandatory)
        :type: str
        :param start: Start sequence number (included in the returned data) (Mandatory)
        :type: int
        :param end:	End sequence number (included in the returned data) (Mandatory)
        :type: int
        :return:
         [
        {
          "symbol": "XBTUSDM",          //Symbol
          "sequence": 1,                //Message sequence number
          "side": "sell",               //Order side
          "orderTime": 1558074650840002300,       //Order placing time
          "size": 10,                   //Order quantity
          "orderId": "5cde551aa14a9cad7e454374", //Order ID
          "price": "7000.0",            //Order price
          "type": "open",                             //Message type
          "clientOid": "xxxxxxxxxx",                  //Optional, this is a user-defined parameter which is used to identify the order
          "ts": 1558074652423004000               //Time at which the order enters the order book- nanosecond
        },
        {
          "symbol": "XBTUSDM",          //Symbol
          "reason": "canceled",     //Reason: canceld or filled
          "sequence": 2,                //Message sequence number
          "orderId": "5cde551aa14a9cad7e454374", //Order ID
          "type": "done",               //Message type
          "ts": 1558075303543002400 //Time at which the order is removed- nanosecond
        }
    ]
        """
        params = {
            'symbol': symbol,
            'start': start,
            'end': end
        }

        return self._request('GET', '/api/v1/level3/message/query', auth=False, params=params)

    def get_ticker(self, symbol):
        """
        https://docs.kumex.com/#get-real-time-ticker
        :param symbol: type tar (Mandatory)
        :return:
        {
          "sequence": 1001,             //Sequence number
          "symbol": "XBTUSDM",              //Symbol
          "side": "buy",                    //Side of liquidity taker
          "size": 10,                       //Filled quantity
          "price": "7000.0",                //Filled price
          "bestBidSize": 20,                //Best bid size
          "bestBidPrice": "7000.0",     //Best bid
          "bestAskSize": 30,                //Best ask size
          "bestAskPrice": "7001.0",     //Best ask
          "tradeId": "5cbd7377a6ffab0c7ba98b26",  //Transaction ID
          "ts": 1550653727731              //Filled time - nanosecond
        }
        """
        params = {
            'symbol': symbol
        }

        return self._request('GET', '/api/v1/ticker', auth=False, params=params)

    def get_contracts_list(self):
        """
        :return:
        {
          "baseCurrency": "XBT",  //Base currency
          "fairMethod": "FundingRate", //Fair price marking method
          "fundingBaseSymbol": ".XBTINT8H",  //Ticker symbol of the based currency
          "fundingQuoteSymbol": ".USDINT8H", //Ticker symbol of the quote currency
          "fundingRateSymbol": ".XBTUSDMFPI8H",  //Funding rate symbol
          "indexSymbol": ".BXBT",    //Index symbol
          "initialMargin": 0.01, //Initial margin requirement
          "isDeleverage": true,   //Enabled ADL or not
          "isInverse": true,  //Reverse contract or not
          "isQuanto": false,   //Whether quanto or not
          "lotSize": 1,   //Minimum lot size
          "maintainMargin": 0.005,    //Maintenance margin requirement
          "makerFeeRate": -0.00025,  //Maker fees
          "makerFixFee": -0.0000000200,   //Fixed maker fees
          "markMethod": "FairPrice", //Marking method
          "maxOrderQty": 1000000,   //Maximum order quantity
          "maxPrice": 1000000.0000000000,  //Maximum order price
          "maxRiskLimit": 200,  //Maximum risk limit (unit: XBT)
          "minRiskLimit": 200,  //Minimum risk limit (unit: XBT)
          "multiplier": -1,    //Contract multiplier
          "quoteCurrency": "USD",  //Quote currency
          "riskStep": 100,  //Risk limit increment value (unit: XBT)
          "rootSymbol": "XBT", //Contract group
          "status": "Open", //Contract status
          "symbol": "XBTUSDM", //Ticker symbol of the contract
          "takerFeeRate": 0.0005,  //Taker fees
          "takerFixFee": 0.0000000600,   //Fixed taker fees
          "tickSize": 1,  //Minimum price changes
          "type": "FFWCSX"    //Type of the contract
        }
        """
        return self._request('GET', '/api/v1/contracts/active', auth=False)

    def get_contract_detail(self, symbol):
        """
        https://docs.kumex.com/#get-order-info-of-the-contract

        :param symbol: type tar (Mandatory)
        :return:
        {
          "baseCurrency": "XBT",  //Base currency
          "fairMethod": "FundingRate", //Fair price marking method
          "fundingBaseSymbol": ".XBTINT8H",  //Ticker symbol of the based currency
          "fundingQuoteSymbol": ".USDINT8H", //Ticker symbol of the quote currency
          "fundingRateSymbol": ".XBTUSDMFPI8H",  //Funding rate symbol
          "indexSymbol": ".BXBT",    //Index symbol
          "initialMargin": 0.01, //Initial margin requirement
          "isDeleverage": true,   //Enabled ADL or not
          "isInverse": true,  //Reverse contract or not
          "isQuanto": false,   //Whether quanto or not
          "lotSize": 1,   //Minimum lot size
          "maintainMargin": 0.005,    //Maintenance margin requirement
          "makerFeeRate": -0.00025,  //Maker fees
          "makerFixFee": -0.0000000200,   //Fixed maker fees
          "markMethod": "FairPrice", //Marking method
          "maxOrderQty": 1000000,   //Maximum order quantity
          "maxPrice": 1000000.0000000000,  //Maximum order price
          "maxRiskLimit": 200,  //Maximum risk limit (unit: XBT)
          "minRiskLimit": 200,  //Minimum risk limit (unit: XBT)
          "multiplier": -1,    //Contract multiplier
          "quoteCurrency": "USD",  //Quote currency
          "riskStep": 100,  //Risk limit increment value (unit: XBT)
          "rootSymbol": "XBT", //Contract group
          "status": "Open", //Contract status
          "symbol": "XBTUSDM", //Ticker symbol of the contract
          "takerFeeRate": 0.0005,  //Taker fees
          "takerFixFee": 0.0000000600,   //Fixed taker fees
          "tickSize": 1,  //Minimum price changes
          "type": "FFWCSX"    //Type of the contract
        }
        """
        return self._request('GET', '/api/v1/contracts/{symbol}'.format(symbol=symbol), auth=False)

    def get_kline_data(self, symbol, granularity:int, begin_t=None, end_t=None):
        """
        https://docs.kucoin.com/futures/#get-k-line-data-of-contract
        :param symbol: type tar (Mandatory)
        :return:
            [
            1575331200000,//时间
            7495.01,      //开盘价
            8309.67,      //最高价
            7250,         //最低价
            7463.55,      //收盘价
            0             //成交量
        ],
        [
            1575374400000,
            7464.37,
            8297.85,
            7273.02,
            7491.44,
            0
        ]
        """
        params = {
            "symbol": symbol,
            "granularity": granularity
        }
        if begin_t:
            params.update({"from": begin_t})
        if end_t:
            params.update({"to": end_t})

        return self._request('GET', '/api/v1/kline/query', auth=False, params=params)

    def get_service_status(self):
        """
        https://docs.kucoin.com/futures/#get-the-service-status
        :param symbol: type tar (Mandatory)
        :return:
        {
        "status": "open",                //open, close, cancelonly
        "msg":  "upgrade match engine"   //remark for operation
        }
        """
        return self._request('GET', '/api/v1/status', auth=False)




class Indicator:

    def __init__(self):
        self.isLocked = False
        self.plannedRecoverTime = 0
        self.indicator = ""
        self.value = 0.0
        self.triggerValue = 0.0

    @staticmethod
    def json_parse(json_data):
        result = Indicator()
        result.isLocked = json_data.get_boolean("isLocked")
        result.plannedRecoverTime = json_data.get_int("plannedRecoverTime")
        result.indicator = json_data.get_float("indicator")
        result.value = json_data.get_float("value")
        result.triggerValue = json_data.get_float("triggerValue")
        return result

class ApiTradingStatus:
    def __init__(self):
        self.indicators = object()

    @staticmethod
    def json_parse(json_data):
        result = ApiTradingStatus()
        
        element_list = list()
        data_list = json_data.get_array("indicators")
        for item in data_list.get_items():
            element = Indicator.json_parse(item)
            element_list.append(element)
        result.positions = element_list

        return result



class PositionMarginHist:

    def __init__(self):
        self.amount = 0.0
        self.asset = ""
        self.symbol = ""
        self.time = 0
        self.type = 0

    @staticmethod
    def json_parse(json_data):
        result = PositionMarginHist()
        result.amount = json_data.get_float("amount")
        result.asset = json_data.get_string("asset")
        result.symbol = json_data.get_string("symbol")
        result.time = json_data.get_int("time")
        result.type = json_data.get_int("type")

        return result



import threading
import websocket
import gzip
import ssl
import logging
from urllib import parse
import urllib.parse

from binance_f.base.printtime import PrintDate
from binance_f.impl.utils.timeservice import get_current_timestamp
from binance_f.impl.utils.urlparamsbuilder import UrlParamsBuilder
from binance_f.impl.utils.apisignature import create_signature
from binance_f.exception.binanceapiexception import BinanceApiException
from binance_f.impl.utils import *
from binance_f.base.printobject import *
from binance_f.model.constant import *
# Key: ws, Value: connection
websocket_connection_handler = dict()


def on_message(ws, message):
    websocket_connection = websocket_connection_handler[ws]
    websocket_connection.on_message(message)
    return


def on_error(ws, error):
    websocket_connection = websocket_connection_handler[ws]
    websocket_connection.on_failure(error)


def on_close(ws):
    websocket_connection = websocket_connection_handler[ws]
    websocket_connection.on_close()


def on_open(ws):
    websocket_connection = websocket_connection_handler[ws]
    websocket_connection.on_open(ws)


connection_id = 0


class ConnectionState:
    IDLE = 0
    CONNECTED = 1
    CLOSED_ON_ERROR = 2


def websocket_func(*args):
    connection_instance = args[0]
    connection_instance.ws = websocket.WebSocketApp(connection_instance.url,
                                                    on_message=on_message,
                                                    on_error=on_error,
                                                    on_close=on_close)
    global websocket_connection_handler
    websocket_connection_handler[connection_instance.ws] = connection_instance
    connection_instance.logger.info("[Sub][" + str(connection_instance.id) + "] Connecting...")
    connection_instance.delay_in_second = -1
    connection_instance.ws.on_open = on_open
    connection_instance.ws.run_forever(sslopt={"cert_reqs": ssl.CERT_NONE})
    connection_instance.logger.info("[Sub][" + str(connection_instance.id) + "] Connection event loop down")
    if connection_instance.state == ConnectionState.CONNECTED:
        connection_instance.state = ConnectionState.IDLE


class WebsocketConnection:

    def __init__(self, api_key, secret_key, uri, watch_dog, request):
        self.__thread = None
        self.url = uri
        self.__api_key = api_key
        self.__secret_key = secret_key
        self.request = request
        self.__watch_dog = watch_dog
        self.delay_in_second = -1
        self.ws = None
        self.last_receive_time = 0
        self.logger = logging.getLogger("binance-futures")
        self.state = ConnectionState.IDLE
        global connection_id
        connection_id += 1
        self.id = connection_id

    def in_delay_connection(self):
        return self.delay_in_second != -1

    def re_connect_in_delay(self, delay_in_second):
        if self.ws is not None:
            self.ws.close()
            self.ws = None
        self.delay_in_second = delay_in_second
        self.logger.warning("[Sub][" + str(self.id) + "] Reconnecting after "
                            + str(self.delay_in_second) + " seconds later")

    def re_connect(self):
        if self.delay_in_second != 0:
            self.delay_in_second -= 1
            self.logger.warning("In delay connection: " + str(self.delay_in_second))
        else:
            self.connect()

    def connect(self):
        if self.state == ConnectionState.CONNECTED:
            self.logger.info("[Sub][" + str(self.id) + "] Already connected")
        else:
            self.__thread = threading.Thread(target=websocket_func, args=[self])
            self.__thread.start()

    def send(self, data):
        self.ws.send(data)

    def close(self):
        self.ws.close()
        del websocket_connection_handler[self.ws]
        self.__watch_dog.on_connection_closed(self)
        self.logger.error("[Sub][" + str(self.id) + "] Closing normally")

    def on_open(self, ws):
        self.logger.info("[Sub][" + str(self.id) + "] Connected to server")
        self.ws = ws
        self.last_receive_time = get_current_timestamp()
        self.state = ConnectionState.CONNECTED
        self.__watch_dog.on_connection_created(self)
        if self.request.subscription_handler is not None:
            self.request.subscription_handler(self)
        return

    def on_error(self, error_message):
        if self.request.error_handler is not None:
            print('error')
            exception = BinanceApiException(BinanceApiException.SUBSCRIPTION_ERROR, error_message)
            self.request.error_handler(exception)
        self.logger.error("[Sub][" + str(self.id) + "] " + str(error_message))

    def on_failure(self, error):
        print('on_failure')
        self.on_error("Unexpected error: " + str(error))
        self.close_on_error()

    def on_message(self, message):
        self.last_receive_time = get_current_timestamp()
        json_wrapper = parse_json_from_string(message)

        if json_wrapper.contain_key("status") and json_wrapper.get_string("status") != "ok":
            error_code = json_wrapper.get_string_or_default("err-code", "Unknown error")
            error_msg = json_wrapper.get_string_or_default("err-msg", "Unknown error")
            self.on_error(error_code + ": " + error_msg)
        elif json_wrapper.contain_key("err-code") and json_wrapper.get_int("err-code") != 0:
            error_code = json_wrapper.get_string_or_default("err-code", "Unknown error")
            error_msg = json_wrapper.get_string_or_default("err-msg", "Unknown error")
            self.on_error(error_code + ": " + error_msg)
        elif json_wrapper.contain_key("result") and json_wrapper.contain_key("id"):
            self.__on_receive_response(json_wrapper)
        else:
            self.__on_receive_payload(json_wrapper)

    def __on_receive_response(self, json_wrapper):
        res = None
        try:
            res = json_wrapper.get_int("id")
        except Exception as e:
            self.on_error("Failed to parse server's response: " + str(e))

        try:
            if self.request.update_callback is not None:
                self.request.update_callback(SubscribeMessageType.RESPONSE, res)
        except Exception as e:
            self.on_error("Process error: " + str(e)
                     + " You should capture the exception in your error handler")

    def __on_receive_payload(self, json_wrapper):
        res = None
        try:
            if self.request.json_parser is not None:
                res = self.request.json_parser(json_wrapper)
        except Exception as e:
            self.on_error("Failed to parse server's response: " + str(e))

        try:
            if self.request.update_callback is not None:
                self.request.update_callback(SubscribeMessageType.PAYLOAD, res)
        except Exception as e:
            self.on_error("Process error: " + str(e)
                     + " You should capture the exception in your error handler")

        if self.request.auto_close:
            self.close()

    def __process_ping_on_trading_line(self, ping_ts):
        self.send("{\"op\":\"pong\",\"ts\":" + str(ping_ts) + "}")
        return

    def __process_ping_on_market_line(self, ping_ts):
        self.send("{\"pong\":" + str(ping_ts) + "}")
        return

    def close_on_error(self):
        if self.ws is not None:
            self.ws.close()
            self.state = ConnectionState.CLOSED_ON_ERROR
            self.logger.error("[Sub][" + str(self.id) + "] Connection is closing due to error")



class RateLimit:

    def __init__(self):
        self.rateLimitType = ""
        self.interval = ""
        self.intervalNum = 0
        self.limit = 0


class ExchangeFilter:

    def __init__(self):
        self.filterType = ""
        self.maxOrders = 0


class Symbol:

    def __init__(self):
        self.symbol = ""
        self.status = ""
        self.onboardDate = 0
        self.maintMarginPercent = 0.0
        self.requiredMarginPercent = 0.0
        self.baseAsset = ""
        self.quoteAsset = ""
        self.pricePrecision = None
        self.quantityPrecision = None
        self.baseAssetPrecision = None
        self.quotePrecision = None
        self.orderTypes = list()
        self.timeInForce = list()
        self.filters = list()


class ExchangeInformation:

    def __init__(self):
        self.timezone = ""
        self.serverTime = 0
        self.rateLimits = list()
        self.exchangeFilters = list()
        self.symbols = list()

    @staticmethod
    def json_parse(json_data):
        result = ExchangeInformation()
        result.timezone = json_data.get_string("timezone")
        result.serverTime = json_data.get_int("serverTime")

        data_list = json_data.get_array("rateLimits")
        element_list = list()
        for item in data_list.get_items():
            element = RateLimit()
            element.rateLimitType = item.get_string("rateLimitType")
            element.interval = item.get_string("interval")
            element.intervalNum = item.get_int("intervalNum")
            element.limit = item.get_int("limit")

            element_list.append(element)
        result.rateLimits = element_list

        data_list = json_data.get_array("exchangeFilters")
        element_list = list()
        for item in data_list.get_items():
            element = ExchangeFilter()
            element.filterType = item.get_string("filterType")
            if element.filterType == "EXCHANGE_MAX_NUM_ORDERS":
                element.maxNumOrders = item.get_int("maxNumOrders")
            elif element.filterType == "EXCHANGE_MAX_ALGO_ORDERS":
                element.maxNumAlgoOrders = item.get_int("maxNumAlgoOrders")

            element_list.append(element)
        result.exchangeFilters = element_list

        data_list = json_data.get_array("symbols")
        element_list = list()
        for item in data_list.get_items():
            element = Symbol()
            element.symbol = item.get_string("symbol")
            element.onboardDate = item.get_int("onboardDate")
            element.status = item.get_string("status")
            element.maintMarginPercent = item.get_float("maintMarginPercent")
            element.requiredMarginPercent = item.get_float("requiredMarginPercent")
            element.baseAsset = item.get_string("baseAsset")
            element.quoteAsset = item.get_string("quoteAsset")
            element.pricePrecision = item.get_int("pricePrecision")
            element.quantityPrecision = item.get_int("quantityPrecision")
            element.baseAssetPrecision = item.get_int("baseAssetPrecision")
            element.quotePrecision = item.get_int("quotePrecision")
            element.orderTypes = item.get_object("orderTypes").convert_2_list()
            element.timeInForce = item.get_object("timeInForce").convert_2_list()

            val_list = item.get_array("filters")
            filter_list = list()
            for jtem in val_list.get_items():
                filter_list.append(jtem.convert_2_dict())
            element.filters = filter_list

            element_list.append(element)
        result.symbols = element_list

        return result



class ListenKeyExpired:
    def __init__(self):
        self.eventType = ""
        self.eventTime = 0
  
    @staticmethod
    def json_parse(json_data):
        result = ListenKeyExpired()
        result.eventType = json_data.get_string("e")
        result.eventTime = json_data.get_int("E")
        
        return result



class Order:

    def __init__(self):
        self.price = 0.0
        self.qty = 0.0


class OrderBook:

    def __init__(self):
        self.lastUpdateId = 0
        self.bids = list()
        self.asks = list()

    @staticmethod
    def json_parse(json_data):
        order_book = OrderBook()
        order_book.lastUpdateId = json_data.get_int("lastUpdateId")

        list_array = json_data.get_array("bids")
        bid_list = list()
        for item in list_array.get_items():
            order = Order()
            val = item.convert_2_list()
            order.price = val[0]
            order.qty = val[1]
            bid_list.append(order)
        order_book.bids = bid_list

        list_array = json_data.get_array("asks")
        ask_list = list()
        for item in list_array.get_items():
            order = Order()
            val = item.convert_2_list()
            order.price = val[0]
            order.qty = val[1]
            ask_list.append(order)
        order_book.asks = ask_list

        return order_book

    def toDict(self):
        return {
            'lastUpdateId': self.lastUpdateId,
            'bids': [[bid.price, bid.qty] for bid in self.bids],
            'asks': [[ask.price, ask.qty] for ask in self.asks]
        }



import json
import logging

import pytest

from TradeGates.TradeGate import TradeGate

loglevel = logging.INFO
logging.basicConfig(level=loglevel)
log = logging.getLogger(__name__)


@pytest.fixture
def get_gates():
    with open("../../config.json") as f:
        config = json.load(f)

    return [TradeGate(config[key], sandbox=True) for key in config.keys()]


def testFullBalance(get_gates):
    for gate in get_gates:
        balance = gate.getBalance()
        print("\nFull Balance from {} exchange: {}".format(gate.exchangeName, balance))
        assert (
            balance is not None
        ), "Problem in fetching balance from {} exchange.".format(gate.exchangeName)

        error_message = "Bad fetch balance interface for {} exchange,".format(
            gate.exchangeName
        )
        try:
            if gate.exchangeName != "Binance":
                if sorted(list(balance[0].keys())) != sorted(
                    ["asset", "free", "locked", "exchangeSpecific"]
                ):
                    assert False, error_message
            else:
                if sorted(list(balance[0].keys())) != sorted(
                    ["asset", "free", "locked"]
                ):
                    assert False, error_message
        except Exception:
            assert False, error_message


def testSingleCoinBalance(get_gates):
    for gate in get_gates:
        balance = gate.get_balance("BTC")
        print(
            "\nSingle coin balance from {} exchange: {}".format(
                gate.exchangeName, balance
            )
        )
        assert (
            balance is not None
        ), "Problem in fetching single coin balance from {} exchange.".format(
            gate.exchangeName
        )

        errorMessage = (
            "Bad fetch single coin balance interface for {} exchange,".format(
                gate.exchangeName
            )
        )
        try:
            if gate.exchangeName != "Binance":
                if sorted(list(balance.keys())) != sorted(
                    ["asset", "free", "locked", "exchangeSpecific"]
                ):
                    assert False, errorMessage
            else:
                if sorted(list(balance.keys())) != sorted(["asset", "free", "locked"]):
                    assert False, errorMessage
        except Exception:
            assert False, errorMessage


def testTradeHistory(get_gates):
    for gate in get_gates:
        if gate.exchangeName.lower() == "kucoin":
            tradeHistory = gate.symbol_account_trade_history("BTC-USDT", futures=False)
        else:
            tradeHistory = gate.symbol_account_trade_history("BTCUSDT", futures=False)
        # print('\nTrade history from {} exchange: {}'.format(gate.exchangeName, tradeHistory))

        assert (
            tradeHistory is not None
        ), "Problem in fetching trade history from {} exchange.".format(
            gate.exchangeName
        )

        interface = [
            "symbol",
            "id",
            "orderId",
            "orderListId",
            "price",
            "qty",
            "quoteQty",
            "commission",
            "commissionAsset",
            "time",
            "isBuyer",
            "isMaker",
            "isBestMatch",
        ]

        errorMessage = "Bad fetch trade history interface for {} exchange,".format(
            gate.exchangeName
        )
        try:
            if not gate.exchangeName == "Binance":
                interface.append("exchangeSpecific")
                if not sorted(list(tradeHistory[0].keys())) == sorted(interface):
                    assert False, errorMessage
            else:
                if not sorted(list(tradeHistory[0].keys())) == sorted(interface):
                    assert False, errorMessage
        except Exception:
            assert False, errorMessage



class AggregateTrade:

    def __init__(self):
        self.id = None
        self.price = 0.0
        self.qty = 0.0
        self.firstId = None
        self.lastId = None
        self.time = 0
        self.isBuyerMaker = False

    @staticmethod
    def json_parse(json_data):
        trade = AggregateTrade()
        trade.id = json_data.get_int("a")
        trade.price = json_data.get_float("p")
        trade.qty = json_data.get_float("q")
        trade.firstId = json_data.get_int("f")
        trade.lastId = json_data.get_int("l")
        trade.time = json_data.get_int("T")
        trade.isBuyerMaker = json_data.get_boolean("m")
        
        return trade


import time
from datetime import datetime

import pandas as pd

from Utils import DataHelpers


def get_balance_out(data, single=False, futures=False):
    if not single:
        out_data = []
        if futures:
            for key, value in data.items():
                coin_data = {
                    "asset": key,
                    "free": value["available_balance"],
                    "locked": value["used_margin"],
                    "exchangeSpecific": value,
                }
                out_data.append(coin_data)
        else:
            for asset in data:
                coin_data = {
                    "asset": asset["coin"],
                    "free": asset["free"],
                    "locked": asset["locked"],
                    "exchangeSpecific": asset,
                }
                out_data.append(coin_data)
    elif futures:
        out_data = set_keys_for_output(data)

    else:
        out_data = {
            "asset": data["coin"],
            "free": data["free"],
            "locked": data["locked"],
            "exchangeSpecific": data,
        }
    return out_data


def set_keys_for_output(data):
    key = list(data.keys())[0]

    return {
        "asset": key,
        "free": data[key]["available_balance"],
        "locked": data[key]["used_margin"],
        "exchangeSpecific": data[key],
    }


def get_my_trade_history_out(data, futures=False):
    out_data = []
    for history in data:
        if futures:
            out_data.append(
                {
                    "symbol": history["symbol"],
                    "id": history["exec_id"],
                    "orderId": history["order_id"],
                    "orderListId": history["order_link_id"],
                    "price": history["price"],
                    "qty": history["order_qty"],
                    "quoteQty": str(
                        float(history["price"]) * float(history["order_qty"])
                    ),
                    "commission": None,
                    "commissionAsset": None,
                    "time": history["trade_time_ms"],
                    "isBuyer": None,
                    "isMaker": None,
                    "isBestMatch": None,
                    "exchangeSpecific": history,
                }
            )
        else:
            out_data.append(
                {
                    "symbol": history["symbol"],
                    "id": history["id"],
                    "orderId": history["orderId"],
                    "orderListId": -1,
                    "price": history["price"],
                    "qty": history["qty"],
                    "quoteQty": str(float(history["price"]) * float(history["qty"])),
                    "commission": history["commission"],
                    "commissionAsset": history["commissionAsset"],
                    "time": history["time"],
                    "isBuyer": history["isBuyer"],
                    "isMaker": history["isMaker"],
                    "isBestMatch": None,
                    "exchangeSpecific": history,
                }
            )
    return out_data


def get_recent_trade_history_out(data, futures=False):
    out_data = []
    for datum in data:
        if futures:
            out_data.append(
                {
                    "id": datum["id"],
                    "price": datum["price"],
                    "qty": datum["qty"],
                    "quoteQty": str(float(datum["qty"] * datum["price"])),
                    "time": datum["trade_time_ms"],
                    "isBuyerMaker": None,
                    "isBestMatch": None,
                    "exchangeSpecific": datum,
                }
            )
        else:
            out_data.append(
                {
                    "id": None,
                    "price": datum["price"],
                    "qty": datum["qty"],
                    "quoteQty": str(float(datum["qty"]) * float(datum["price"])),
                    "time": datum["time"],
                    "isBuyerMaker": datum["isBuyerMaker"],
                    "isBestMatch": None,
                    "exchangeSpecific": datum,
                }
            )
    return out_data


def get_make_spot_order_out(data):
    return {
        "symbol": data["symbol"],
        "orderId": data["orderId"],
        "orderListId": -1,
        "clientOrderId": data["orderLinkId"],
        "transactTime": data["transactTime"],
        "price": data["price"],
        "origQty": data["origQty"],
        "executedQty": data["executedQty"],
        "cummulativeQuoteQty": None,
        "status": data["status"],
        "timeInForce": data["timeInForce"],
        "type": data["type"],
        "side": data["side"],
        "fills": None,
        "exchangeSpecific": data,
    }


def get_order_out(data, futures=False):
    if not futures:
        return {
            "symbol": data["symbol"],
            "orderId": data["orderId"],
            "orderListId": -1,
            "clientOrderId": data["orderLinkId"],
            "price": data["price"],
            "origQty": data["origQty"],
            "executedQty": data["executedQty"],
            "cummulativeQuoteQty": data["cummulativeQuoteQty"],
            "status": data["status"],
            "timeInForce": data["timeInForce"],
            "type": data["type"],
            "side": data["side"],
            "stopPrice": data["stopPrice"],
            "icebergQty": data["icebergQty"],
            "time": data["time"],
            "updateTime": data["updateTime"],
            "isWorking": data["isWorking"],
            "origQuoteOrderQty": None,
            "exchangeSpecific": data,
        }


def get_open_orders_out(data, futures=False):
    out_data = []
    if not futures:
        out_data.extend(
            {
                "symbol": datum["symbol"],
                "orderId": datum["orderId"],
                "orderListId": None,
                "clientOrderId": datum["orderLinkId"],
                "price": datum["price"],
                "origQty": datum["origQty"],
                "executedQty": datum["executedQty"],
                "cummulativeQuoteQty": datum["cummulativeQuoteQty"],
                "status": datum["status"],
                "timeInForce": datum["timeInForce"],
                "type": datum["type"],
                "side": datum["side"],
                "stopPrice": datum["stopPrice"],
                "icebergQty": datum["icebergQty"],
                "time": datum["time"],
                "updateTime": datum["updateTime"],
                "isWorking": datum["isWorking"],
                "origQuoteOrderQty": None,
                "exchangeSpecific": datum,
            }
            for datum in data
        )
    return out_data


def futures_order_out(data, is_conditional=False):
    if is_conditional:
        return {
            "symbol": data["symbol"],
            "orderId": data["stop_order_id"],
            "clientOrderId": data["order_link_id"],
            "transactTime": time.mktime(
                datetime.strptime(
                    data["created_time"], "%Y-%m-%dT%H:%M:%SZ"
                ).timetuple()
            ),
            "price": data["price"],
            "origQty": data["qty"],
            "executedQty": 0.0,
            "cummulativeQuoteQty": 0.0,
            "status": data["order_status"],
            "timeInForce": data["time_in_force"],
            "type": data["order_type"],
            "side": data["side"],
            "extraData": {
                "reduceOnly": data["reduce_only"],
                "stopPrice": data["trigger_price"],
                "workingType": data["trigger_by"],
                "avgPrice": 0.0,
                "origType": data["order_type"],
                "positionSide": None,
                "activatePrice": None,
                "priceRate": None,
                "closePosition": data["close_on_trigger"],
            },
            "exchangeSpecific": data,
        }
    else:
        return {
            "symbol": data["symbol"],
            "orderId": data["order_id"],
            "clientOrderId": data["order_link_id"],
            "transactTime": time.mktime(
                datetime.strptime(
                    data["created_time"], "%Y-%m-%dT%H:%M:%SZ"
                ).timetuple()
            ),
            "price": data["price"],
            "origQty": data["qty"],
            "executedQty": data["cum_exec_qty"],
            "cummulativeQuoteQty": data["cum_exec_value"],
            "status": data["order_status"],
            "timeInForce": data["time_in_force"],
            "type": data["order_type"],
            "side": data["side"],
            "extraData": {
                "reduceOnly": data["reduce_only"],
                "stopPrice": 0.0,
                "workingType": None,
                "avgPrice": 0.0,
                "origType": data["order_type"],
                "positionSide": None,
                "activatePrice": None,
                "priceRate": None,
                "closePosition": data["close_on_trigger"],
            },
            "exchangeSpecific": data,
        }


def make_dummy_balance(asset):
    return {
        "asset": asset,
        "free": str(0.0),
        "locked": str(0.0),
        "exchangeSpecific": {},
    }


def get_spot_order_as_dict(order: DataHelpers.OrderData):
    return {
        "symbol": order.symbol,
        "qty": order.quantity,
        "side": order.side,
        "type": order.order_type,
        "timeInForce": order.time_in_force,
        "price": order.price,
        "orderLinkId": order.new_client_order_id,
    }


def get_futures_order_as_dict(
    order: DataHelpers.FuturesOrderData, time_in_force_translate
):
    if "STOP" in order.orderType:
        params = {
            "side": order.side.lower().title(),
            "symbol": order.symbol,
            "order_type": "Market" if order.orderType == "STOP_MARKET" else "Limit",
            "qty": order.quantity,
            "price": order.price,
            "base_price": order.extra_params["basePrice"],
            "stop_px": order.stop_price,
            "time_in_force": order.time_in_force
            if order.time_in_force in time_in_force_translate.values()
            else time_in_force_translate[order.time_in_force],
            "close_on_trigger": order.close_position,
            "reduce_only": order.reduce_only,
        }

        if "triggerBy" in order.extra_params.keys():
            params["trigger_by"] = order.extra_params["triggerBy"]

    else:
        params = {
            "side": order.side.lower().title(),
            "symbol": order.symbol,
            "order_type": order.orderType.lower().title(),
            "qty": order.quantity,
            "time_in_force": order.time_in_force
            if order.time_in_force in time_in_force_translate.values()
            else time_in_force_translate[order.time_in_force],
            "close_on_trigger": order.close_position,
            "reduce_only": order.reduce_only,
        }

    if order.price is not None:
        params["price"] = order.price

    if order.new_client_order_id is not None:
        params["order_link_id"] = order.new_client_order_id

    if "takeProfit" in order.extra_params.keys():
        params["take_profit"] = order.extra_params["takeProfit"]

    if "stopLoss" in order.extra_params.keys():
        params["stop_loss"] = order.extra_params["stopLoss"]

    if "tpTriggerBy" in order.extra_params.keys():
        params["tp_trigger_by"] = order.extra_params["tpTriggerBy"]

    if "slTriggerBy" in order.extra_params.keys():
        params["sl_trigger_by"] = order.extra_params["slTriggerBy"]

    if "positionIdx" in order.extra_params.keys():
        params["position_idx"] = order.extra_params["positionIdx"]

    return params


def convert_interval_to_futures_klines(interval):
    if interval == "1m":
        return 1
    elif interval == "3m":
        return 3
    elif interval == "5m":
        return 5
    elif interval == "15m":
        return 15
    elif interval == "30m":
        return 30
    elif interval == "1h":
        return 60
    elif interval == "2h":
        return 120
    elif interval == "4h":
        return 240
    elif interval == "6h":
        return 360
    elif interval == "12h":
        return 720
    elif interval == "1d":
        return "D"
    elif interval == "1w":
        return "W"
    elif interval == "1M":
        return "M"


def get_interval_in_seconds(interval, valid_intervals):
    if interval not in valid_intervals:
        raise ValueError("Incorrect time interval specified")
    if interval == "1m":
        return 60
    elif interval == "3m":
        return 3 * 60
    elif interval == "5m":
        return 5 * 60
    elif interval == "15m":
        return 15 * 60
    elif interval == "30m":
        return 30 * 60
    elif interval == "1h":
        return 60 * 60
    elif interval == "2h":
        return 120 * 60
    elif interval == "4h":
        return 240 * 60
    elif interval == "6h":
        return 360 * 60
    elif interval == "12h":
        return 720 * 60
    elif interval == "1d":
        return 86400
    elif interval == "1w":
        return 7 * 86400
    elif interval == "1M":
        return 30 * 86400


def get_klines_desired_only_cols(data, desired_indexes):
    return [[datum[index] for index in desired_indexes] for datum in data]


def klines_convert_to_pandas(final_data_array):
    df = pd.DataFrame(
        final_data_array,
        columns=[
            "date",
            "open",
            "high",
            "low",
            "close",
            "volume",
            "closeDate",
            "tradesNum",
        ],
    )
    df.set_index("date", inplace=True)
    return df


def klines_convert_date(data, futures, time_col_idxs):
    for datum in data:
        for idx in time_col_idxs:
            if futures:
                datum[idx] = datetime.fromtimestamp(float(datum[idx]))
            else:
                datum[idx] = datetime.fromtimestamp(float(datum[idx]) / 1000)



from binance_f.model.constant import *
from binance_f.model.message import Msg
from binance_f.model.exchangeinformation import ExchangeInformation
from binance_f.model.orderbook import OrderBook
from binance_f.model.trade import Trade
from binance_f.model.aggregatetrade import AggregateTrade
from binance_f.model.candlestick import Candlestick
from binance_f.model.markprice import MarkPrice
from binance_f.model.openinterest import OpenInterest
from binance_f.model.fundingrate import FundingRate
from binance_f.model.tickerpricechangestatistics import TickerPriceChangeStatistics
from binance_f.model.symbolprice import SymbolPrice
from binance_f.model.symbolorderbook import SymbolOrderBook
from binance_f.model.liquidationorder import LiquidationOrder
from binance_f.model.aggregatetradeevent import AggregateTradeEvent
from binance_f.model.markpriceevent import MarkPriceEvent
from binance_f.model.candlestickevent import CandlestickEvent
from binance_f.model.symbolminitickerevent import SymbolMiniTickerEvent
from binance_f.model.symboltickerevent import SymbolTickerEvent
from binance_f.model.symbolbooktickerevent import SymbolBookTickerEvent
from binance_f.model.liquidationorderevent import LiquidationOrderEvent
from binance_f.model.orderbookevent import OrderBookEvent
from binance_f.model.diffdepthevent import DiffDepthEvent
from binance_f.model.order import Order
from binance_f.model.balance import Balance
from binance_f.model.accountinformation import AccountInformation
from binance_f.model.leverage import Leverage
from binance_f.model.codeandmsg import CodeMsg
from binance_f.model.positionmode import PositionMode
from binance_f.model.positionmargin import PositionMargin
from binance_f.model.positionmarginhistory import PositionMarginHist
from binance_f.model.position import Position
from binance_f.model.mytrade import MyTrade
from binance_f.model.income import Income
from binance_f.model.accountupdate import AccountUpdate
from binance_f.model.orderupdate import OrderUpdate
from binance_f.model.listenkeyexpired import ListenKeyExpired

from binance_f.model.accountinformationv2 import AccountInformationV2
from binance_f.model.adlquantile import AdlQuantile
from binance_f.model.apitradingstatus import ApiTradingStatus
from binance_f.model.balancev2 import BalanceV2
from binance_f.model.blvtnavcandlestick import BLVTNAVCandlestick
from binance_f.model.countdowncancelall import CountdownCancelAll
from binance_f.model.indexInfo import IndexInfo
from binance_f.model.leveragebracket import LeverageBracket
from binance_f.model.openintereststats import OpenInterestStats
from binance_f.model.longshortratio import LongShortRatio
from binance_f.model.takerbuysellratio import TakerBuySellRatio
from binance_f.model.blvtinfoevent import BLVTInfoEvent
from binance_f.model.blvtnavcandlestickevent import BLVTNAVCandlestickEvent
from binance_f.model.compositeindexevent import CompositeIndexEvent